# -*- coding: utf-8 -*-

import logging
import time


class Cluster:
    @classmethod
    def start_cluster(cls, nworkers: int, sleep_step: float = 5):
        logging.info(f"Starting cluster with {nworkers} workers")

        # on start-up we need to submit the target number of jobs
        job_ids = cls.submit_jobs(nworkers)

        # we now monitor the jobs running and submit new workers whenever
        # we drop below out target.
        # We do this endlessly until the user closes the script
        while True:
            job_ids = cls.update_jobs_list(job_ids)
            njobs_needed = nworkers - len(job_ids)
            if njobs_needed > 0:
                job_ids += cls.submit_jobs(njobs_needed)
            time.sleep(sleep_step)

    @classmethod
    def wait_for_jobs(cls, job_ids: list[int], sleep_step: float = 5):
        # loop until the job id list is empty
        while job_ids:
            job_ids = cls.update_jobs_list(job_ids)
            time.sleep(sleep_step)

    @classmethod
    def submit_jobs(cls, njobs: int) -> list[int]:
        """
        Calls submit_to_queue a set number of times and returns the new job ids
        """
        job_ids = [cls.submit_job() for n in range(njobs)]
        logging.info(f"{njobs} new workers have been submitted")
        return job_ids

    @staticmethod
    def submit_job() -> int:
        """
        Submits a new job to the queue and returns the job id
        """
        raise NotImplementedError(
            "add a custom submit_job method to your cluster class"
        )

    @staticmethod
    def update_jobs_list(job_ids: list[int]) -> int:
        """
        Given a list of job ids, it will check which ones are still running and
        which are finished. It will then return a list of the job id that are
        still running.
        """
        raise NotImplementedError(
            "add a custom update_jobs_list method to your cluster class"
        )



# -*- coding: utf-8 -*-

"""
This defines the base "simmate-vasp" command that all other commands stem from.
"""

import typer

from simmate.apps.vasp.command_line.inputs import inputs_app
from simmate.apps.vasp.command_line.plot import plot_app

vasp_app = typer.Typer(rich_markup_mode="markdown")


@vasp_app.callback(no_args_is_help=True)
def base_command():
    """
    A collection of utilities to help work with VASP
    """
    # When we call the command "simmate-vasp" this is where we start, and it then
    # looks for all other functions that have the decorator "@simmate_vasp.command()"
    # to decide what to do from there.
    pass


# All commands are organized into other files, so we need to manually register
# them to our base "simmate-vasp" command here.
vasp_app.add_typer(inputs_app, name="inputs")
vasp_app.add_typer(plot_app, name="plot")



# -*- coding: utf-8 -*-

from pymatgen.analysis.structure_matcher import StructureMatcher

from simmate.toolkit import Composition, Structure
from simmate.toolkit.creators.structure.prototypes import FromAflowPrototypes


def get_structures_from_prototypes(
    composition: Composition,
    remove_matching: bool = True,
    **kwargs,
) -> list[Structure]:
    """
    Generates a list of structures by using prototypes as templates. This function
    is a simple wrapper around the functionality in
    `simmate.toolkit.creators.structures.prototypes`. For more control over the
    structures generated, refer to the underlying creator class.
    """

    structures = FromAflowPrototypes(
        composition,
        **kwargs,
    ).structures

    if remove_matching:
        matcher = StructureMatcher()
        groups = matcher.group_structures(structures)
        # just grab the first result of each group
        structures = [group[0] for group in groups]

    return structures



# -*- coding: utf-8 -*-

import base64
from io import BytesIO
from pathlib import Path

from django.utils.safestring import mark_safe

from simmate.visualization.plotting import Figure


class MatplotlibFigure(Figure):
    @classmethod
    def view_plot(cls, parent_class, **kwargs):
        figure = cls.get_plot(parent_class, **kwargs)
        figure.show()

    @classmethod
    def write_plot(cls, parent_class, directory: Path = None, **kwargs):
        if not directory:
            directory = Path.cwd()
        figure = cls.get_plot(parent_class, **kwargs)
        filename = directory / f"{cls.name}.png"
        figure.savefig(filename)

    @classmethod
    def get_html_div(cls, parent_class, **kwargs):
        figure = cls.get_plot(parent_class, **kwargs)

        # Lmao I have no idea how this works and I just got lucky when following
        # these stack overflow posts... I should switch to plotly ASAP so I can
        # remove this filter.
        #   https://stackoverflow.com/questions/14824522/
        #   https://stackoverflow.com/questions/40534715/
        # Alternatively, I could rewrite this function to save a randomly generated
        # file to the /static/runtime folder and then passing back an html <img> tag
        # that points back to this image.

        figdata = BytesIO()
        figure.savefig(figdata, format="png")
        figdata.seek(0)
        data = figdata.getvalue()
        data_decoded = base64.b64encode(data).decode("utf-8").replace("\n", "")
        html = f"""
        <img id="ItemPreview" class="img-fluid" src="data:image/png;base64,{data_decoded}">
        """

        # BUG: every time this filter is called, I recieve the following warning:
        #
        # .../lib/python3.10/site-packages/pymatgen/util/plotting.py:48: UserWarning:
        # Starting a Matplotlib GUI outside of the main thread will likely fail.
        # plt.figure(figsize=(width, height), facecolor="w", dpi=dpi)
        #
        # This doesn't seem to cause any issues, so I ignore it for now.

        # Because we added new html to our script, we need to have Django check it
        # ensure it safe before returning. Read more about this here:
        # https://docs.djangoproject.com/en/3.2/howto/custom-template-tags/#filters-and-auto-escaping
        return mark_safe(html)



# -*- coding: utf-8 -*-

from django.contrib import admin
from django.urls import include, path

from simmate.configuration import settings
from simmate.utilities import get_app_submodule
from simmate.website.core import views


def get_app_urls():
    # First let's find our custom simmate apps that supply a 'urls' module. We
    # want to automatically make these avaialable in the web UI.

    extra_url_paths = []

    for app_name in settings.apps:
        urls_path = get_app_submodule(app_name, "urls")
        if urls_path:
            simple_name = urls_path.split(".")[-2]
            new_path = path(
                route=f"apps/{simple_name}/",
                # set the namespace so that we can easily look up app urls
                #   https://stackoverflow.com/questions/48608894
                view=include((urls_path, simple_name), namespace=simple_name),
                name=simple_name,
            )
            extra_url_paths.append(new_path)

    return extra_url_paths


def get_disabled_urls():
    """
    Collects any pages that should be disabled
    """
    disabled_url_paths = []

    # We want to turn off the new account signup form if we require all users
    # to sign in using their allauth (e.g. Microsoft login)
    if settings.website.require_login_internal:
        new_path = path(route="accounts/signup/", view=views.permission_denied)
        disabled_url_paths.append(new_path)

    return disabled_url_paths


urlpatterns = [
    #
    # This is the path to the homepage (just simmate.org)
    path(route="", view=views.home, name="home"),
    #
    # Disabled urls (such as the account signup form), must come first. The only
    # page that can't be disabled is the home page.
    *get_disabled_urls(),
    #
    # This is the built-in admin site that django provides
    path(route="admin/", view=admin.site.urls, name="admin"),
    #
    # This is the profile system with login/logout.
    path(
        route="accounts/",
        view=include("allauth.urls"),
        name="accounts",
    ),
    # On login success, you will be pointed to /accounts/profile by default.
    # If you want to change this defualt, then set LOGIN_REDIRECT_URL in
    # the settings.py file
    path(route="accounts/profile/", view=views.profile, name="profile"),
    # When you sign out, you are sent to LOGOUT_REDIRECT_URL (set in settings.py)
    path(route="accounts/loginstatus/", view=views.loginstatus, name="loginstatus"),
    #
    #
    path(
        route="data/",
        view=include(
            ("simmate.website.data_explorer.urls", "simmate.website.data_explorer"),
            namespace="data_explorer",
        ),
        name="data_explorer",
    ),
    #
    #
    # All local calculations are stored at this endpoint
    path(
        route="workflows/",
        view=include("simmate.website.workflows.urls"),
        name="workflows",
    ),
    #
    # This app includes core functionality, such as views for crystal structures
    # in a 3D viewport.
    path(
        route="core-components/",
        view=include(
            (
                "simmate.website.core_components.urls",
                "simmate.website.core_components",
            ),
            namespace="core_components",
        ),
        name="core_components",
    ),
    #
    # And extra one-page views
    path(route="extras/", view=views.extras, name="extras"),
    path(route="contact/", view=views.contact, name="contact"),
    path(route="about/", view=views.about, name="about"),
    #
    # Django-unicorn urls
    path("unicorn/", include("django_unicorn.urls")),
    #
    # Django-contrib-comments urls
    path("comments/", include("django_comments.urls")),
    #
    # Custom Simmate apps (if present)
    path(route="apps/", view=views.apps, name="apps"),
    *get_app_urls(),
]



# -*- coding: utf-8 -*-

from simmate.database import connect

from .badelf import (
    BadElf__Badelf__BadelfHse,
    BadElf__Badelf__BadelfPbesol,
    BadElf__Badelf__BadelfTest,
    StaticEnergy__Vasp__WarrenLabPrebadelfHse,
    StaticEnergy__Vasp__WarrenLabPrebadelfPbesol,
)
from .nested_dft import (
    Nested__Vasp__WarrenLabRelaxationStaticHseHse,
    Nested__Vasp__WarrenLabRelaxationStaticPbeHse,
    Nested__Vasp__WarrenLabRelaxationStaticPbePbe,
    Relaxation__Vasp__WarrenLabHseWithWavecar,
    Relaxation__Vasp__WarrenLabPbeWithWavecar,
)
from .relaxation import (
    Relaxation__Vasp__WarrenLabHse,
    Relaxation__Vasp__WarrenLabHsesol,
    Relaxation__Vasp__WarrenLabPbe,
    Relaxation__Vasp__WarrenLabPbeMetal,
    Relaxation__Vasp__WarrenLabPbesol,
    Relaxation__Vasp__WarrenLabScan,
)
from .static_energy import (
    StaticEnergy__Vasp__WarrenLabHse,
    StaticEnergy__Vasp__WarrenLabHsesol,
    StaticEnergy__Vasp__WarrenLabPbe,
    StaticEnergy__Vasp__WarrenLabPbeMetal,
    StaticEnergy__Vasp__WarrenLabPbesol,
    StaticEnergy__Vasp__WarrenLabScan,
)



# -*- coding: utf-8 -*-

import shutil
import time
from pathlib import Path
from tempfile import mkdtemp


def get_directory(directory: Path | str = None) -> Path:
    """
    Initializes a directory.

    There are many cases where the user can choose their working directory
    for a calculation, and they may want to provide their directory in various
    formats. This includes... None, a string, or a TemporaryDirectory instance.
    Based on the input, this function does the following:
      - `None`:
          returns the full path to a new folder inside python's
          current working directory named "simmate-task-<randomID>"
      - `str` or `pathlib.Path`:
          makes the directory if it doesnt exist and then returns the full path

    #### Parameters

    - `directory`:
        Either None, a path to the directory, or a tempdir. The default is None.

    #### Returns

    - `directory`:
        The full path to the initialized directory as a `pathlib.Path` object
    """

    # if no directory was provided, we create a new folder within the current
    # working directory. All of these folders are named randomly.
    # Note: we can't name these nicely (like simmate-task-001) because that will
    # introduce race conditions when making these folders in production.
    if not directory:
        # create a directory in the current working directory. Note, even though
        # we are creating a "TemporaryDirectory" here, this directory is never
        # actually deleted. Note, Path() gives the working directory
        directory_new = mkdtemp(prefix="simmate-task-", dir=Path.cwd())
        directory_cleaned = Path(directory_new)

    # otherwise make sure the directory the user provided exists and if it does
    # not, then make it!
    else:
        directory_cleaned = Path(directory)
        if not directory_cleaned.exists():
            # We use parents=True because we want to make these directories
            # recursively. That is, we can do "path/to/newfolder1/newfolder2" where
            # multiple folders can be made with one call.
            # Also if the folder already exists, we don't want to raise an error.
            directory_cleaned.mkdir(parents=True, exist_ok=True)

    # and return the full path to the directory
    return directory_cleaned.absolute()


def copy_files_from_directory(
    files_to_copy: list[Path],
    directory_old: Path,
    directory_new: Path = None,
) -> Path:
    """
    Given an old directory, copies all of the files listed over to a new one.

    #### Parameters

    - `files_to_copy`:
        a list of filenames that should be copied over. If you want ALL files,
        try using the `copy_directory` utility instead

    - `directory_old`:
        Name of the directory to be copied over. Must exist.

    - `directory_new`:
        Name of the new directory (optional). This will be passed to the
        `get_directory` utility.

    #### Returns

    - `directory`:
        The path to the new directory as a string

    """

    # Start by creating a new directory or grabbing the one given.
    directory_new_cleaned = get_directory(directory_new)

    # and make sure we have a path obj
    directory_old = Path(directory_old)

    # TODO: this is a copy/paste of the code in copy_directory. Consider
    # refactoring and making a context manager (bc the delete_tmp is used below)
    # First check if the previous directory exists. There are several
    # possibilities that we need to check for:
    #   1. directory exists on the same file system and can be found
    #   2. directory exists on the same file system but is now an archive
    #   3. directory/archive is on another file system (requires ssh to access)
    #   4. directory was deleted and unavailable
    if directory_old.exists():
        # everything is good to go
        delete_temp = False
    elif directory_old.with_suffix(".zip").exists():
        # unpack the old archive
        shutil.unpack_archive(
            filename=directory_old.with_suffix(".zip"),
            extract_dir=directory_old.parent,
        )
        delete_temp = True
    else:
        raise Exception(
            "Unable to locate the previous directory to copy. Make sure the "
            "past directory is located on the same file system. Directory that "
            f"couldn't be found was... {directory_old}"
        )
    # TODO: for possibility 3, I could implement automatic copying with
    # the "fabric" python package (uses ssh). I'd also need to store
    # filesystem names (e.g. "WarWulf") to know where to connect.

    for filename in files_to_copy:
        source_file = directory_old / filename
        destination = directory_new_cleaned / filename
        if not source_file.exists:
            raise Exception(
                "Unable to locate the file to copy. Make sure the "
                "past directory is located on the same file system. File that "
                f"couldn't be found was... {source_file}"
            )
        shutil.copy(source_file, destination)

    # Then remove the unpacked archive now that we copied it.
    # This leaves the original archive behind and unaltered too.
    if delete_temp:
        shutil.rmtree(directory_old)

    return directory_new_cleaned


def copy_directory(
    directory_old: Path,
    directory_new: Path = None,
    ignore_simmate_files: bool = False,
) -> Path:
    """
    Given an old directory, copies all of it's contents over to a new one.
    Optionally, you can avoid copying any simmate_* files and archives within
    that original directory to save on disk space.

    #### Parameters

    - `directory_old`:
        Name of the directory to be copied over. Must exist.

    - `directory_new`:
        Name of the new directory (optional). This will be passed to the
        `get_directory` utility.

    - `ignore_simmate_files`:
        Whether to ignore simmate_* files when copying over. Defaults to False.

    #### Returns

    - `directory`:
        The path to the new directory as a string

    """

    # Start by creating a new directory or grabbing the one given.
    directory_new_cleaned = get_directory(directory_new)

    # and make sure we have a path obj
    directory_old = Path(directory_old)

    # First check if the previous directory exists. There are several
    # possibilities that we need to check for:
    #   1. directory exists on the same file system and can be found
    #   2. directory exists on the same file system but is now an archive
    #   3. directory/archive is on another file system (requires ssh to access)
    #   4. directory was deleted and unavailable
    # When copying over the directory, we ignore any `simmate_` files
    # that correspond to metadata/results/corrections/etc.
    if directory_old.exists():
        # copy the old directory to the new one
        shutil.copytree(
            src=directory_old,
            dst=directory_new_cleaned,
            ignore=(
                shutil.ignore_patterns("simmate_*") if ignore_simmate_files else None
            ),
            dirs_exist_ok=True,
        )
    elif directory_old.with_suffix(".zip").exists():
        # unpack the old archive
        shutil.unpack_archive(
            filename=directory_old.with_suffix(".zip"),
            extract_dir=directory_old.parent,
        )
        # copy the old directory to the new one
        shutil.copytree(
            src=directory_old,
            dst=directory_new_cleaned,
            ignore=(
                shutil.ignore_patterns("simmate_*") if ignore_simmate_files else None
            ),
            dirs_exist_ok=True,
        )
        # Then remove the unpacked archive now that we copied it.
        # This leaves the original archive behind and unaltered too.
        shutil.rmtree(directory_old)
    else:
        raise Exception(
            "Unable to locate the previous directory to copy. Make sure the "
            "past directory is located on the same file system. Directory that "
            f"couldn't be found was... {directory_old}"
        )
    # TODO: for possibility 3, I could implement automatic copying with
    # the "fabric" python package (uses ssh). I'd also need to store
    # filesystem names (e.g. "WarWulf") to know where to connect.

    return directory_new_cleaned


def make_archive(directory: Path, files_to_exclude: list[str] = []):
    """
    Compresses the directory to a zip file of the same name. After compressing,
    it then deletes the original directory.

    #### Parameters

    - `directory`:
        Path to the folder that should be archived
    """

    directory_full = directory.absolute()

    # Remove any files that were requested to be deleted. For example, POTCAR
    # files of VASP calculations.
    for file_to_remove in files_to_exclude:
        for file_found in directory.rglob(file_to_remove):
            file_found.unlink()

    # This wraps shutil.make_archive to change the default parameters. Normally,
    # it writes the archive in the working directory, but we update it to use the
    # the same directory as the folder being archived. The format is also set
    # to zip.
    shutil.make_archive(
        # By default I choose within the current directory and save
        # it as the same name of the directory (+ zip ending)
        base_name=directory_full,
        # format to use switch to gztar after testing
        format="zip",
        # full path to up tp directory that will be archived
        root_dir=directory_full.parent,
        # directory within root_directory to archive
        base_dir=directory_full.name,
    )
    # now remove the directory we just archived
    shutil.rmtree(directory)


def make_error_archive(directory: Path):
    """
    Compresses the directory to a zip file and stores the new archive within the
    original. This utility is meant for creating archives within the directory
    of a failed calculation, so the new archive will be named something like
    `simmate_attempt_01.zip`, where the number is automatically determined. When
    archiving the folder, all "simmate_*" files within the directory are ignore
    (this is includes earlier simmate_attempt_*.zip archives).

    #### Parameters

    - `directory`:
        Path to the folder that should be archived
    """

    full_path = directory.absolute()

    # check the directory and see how many other "simmate_attempt_*.zip" files
    # already exist. Our archive number will be based off of this.
    count = (
        len([f for f in full_path.iterdir() if f.name.startswith("simmate_attempt_")])
        + 1
    )
    count_str = str(count).zfill(2)
    base_name = full_path / f"simmate_attempt_{count_str}"

    # Before we make the archive, we want to avoid also storing other simmate
    # archives and files within this new archive. We therefore copy all files
    # that do NOT start with "simmate_" over into a folder and then archive it.
    shutil.copytree(
        src=full_path,
        dst=base_name,
        ignore=shutil.ignore_patterns("simmate_*"),
    )

    # now convert the copied files to a new archive
    make_archive(base_name)


def archive_old_runs(
    directory: Path = None,
    time_cutoff: float = 3 * 7 * 24 * 60 * 60,  # equal to 3 weeks
):
    """
    Goes through a given directory and finds all "simmate-task-" folders that
    are older than a given time cutoff. Each of these folders is then compressed
    to a zip file and then the original folder is removed.

    #### Parameters

    - `directory`:
        base directory that will contain folders to archive. Defaults to the
        working directory.
    - `time_cutoff`:
        The time (in seconds) required to determine whether a folder is old or not.
        If the folder is considered old, then it will be archived and then deleted.
        The default is 3 weeks.

    """
    if not directory:
        directory = Path.cwd()

    # load the full path to the desired directory
    directory = get_directory(directory)

    # grab all files/folders in this directory and then limit this list to those
    # that are...
    #   1. folders
    #   2. start with "simmate-task-"
    #   3. haven't been modified for at least time_cutoff
    foldernames = []
    for foldername in directory.iterdir():
        foldername_full = directory / foldername
        if (
            foldername_full.is_dir()
            and "simmate-task-" in foldername.name
            and time.time() - foldername_full.lstat().st_mtime > time_cutoff
        ):
            foldernames.append(foldername_full)

    # now go through this list and archive the folders that met the criteria
    [make_archive(f) for f in foldernames]


def empty_directory(directory: Path, files_to_keep: list[Path] = []):
    """
    Deletes all files and folders within a directory, except for those provided
    to the files_to_keep parameter.

    #### Parameters

    - `directory`:
        base directory that should be emptied
    - `files_to_keep`:
        A list of file and folder names within the base directory that should
        not be deleted. The default is [].
    """
    # grab all of the files and folders inside the listed directory
    for filename in directory.iterdir():
        full_path = filename.absolute()
        if filename not in files_to_keep:
            # check if we have a folder or a file.
            # Folders we delete with shutil and files with the os module
            if full_path.is_dir():
                shutil.rmtree(full_path)  # ignore_errors=False
            else:
                full_path.unlink()


def chunk_read(
    filename: Path | str,
    chunk_size: int,
    delimiter: str,
) -> list:
    """
    Yields n-sized chunks of entries from a file, where the chunks sizes are
    determined by a delimiter.

    This is useful for reading large files, such as a molecular format or CSV
    that is too large to load into memory.

    As an alternative to this function, consider using Dask to load larger than
    memory CSVs.
    """
    filename = Path(filename)
    with filename.open("r") as file:
        is_end = False
        while not is_end:
            delimiter_count = 1  # start at 1 bc we need to match chunk_size at end
            lines = []
            while delimiter_count < chunk_size:
                line = file.readline()
                if not line:
                    is_end = True
                    break
                lines.append(line)
                if delimiter in line:
                    delimiter_count += 1
            lines = "".join(lines)
            yield lines.split(delimiter)



# -*- coding: utf-8 -*-


import subprocess
from pathlib import Path
from tempfile import mkstemp

from simmate.engine.execution.cluster.base import Cluster


class LocalCluster(Cluster):
    """
    Submits workers via subprocesses
    """

    worker_command: str = "simmate engine start-worker"

    @classmethod
    def submit_job(cls) -> subprocess.Popen:
        output_file = (
            Path.cwd()
            / mkstemp(
                prefix="simmate-worker-",
                suffix=".out",
                dir=Path.cwd(),
            )[1]
        )

        popen = subprocess.Popen(
            cls.worker_command,
            shell=True,
            stdout=output_file.open("w"),
            stderr=output_file.open("w"),
        )
        return popen

    @staticmethod
    def update_jobs_list(job_ids: list[subprocess.Popen]) -> list[subprocess.Popen]:
        # each job id is actually a subprocess.Popen object
        still_running = []
        for process in job_ids:
            if process.poll() is None:
                # p.subprocess is alive
                still_running.append(process)

        return still_running



# -*- coding: utf-8 -*-

import math

from pymatgen.core import Lattice

from simmate.toolkit import Structure


class Kpoints:
    # placeholder class until we have general one for both vasp, qe, & others

    def __init__(
        self,
        mode: str,
        grid: list[list[float]] | list[float],
        offset: list[list[float]] | list[float],
        weights: list[list[float]] | list[float],
    ):
        self.mode = mode
        self.grid = grid
        self.offset = offset
        self.weights = weights
        # TODO: maybe add spacing and density attribute for REAL values
        # then original_spacing and original_density as well

    def to_dict(self) -> dict:
        return {
            "mode": self.mode,
            "grid": self.grid,
            "offset": self.offset,
            "weights": self.weights,
        }

    @classmethod
    def from_spacing(
        cls,
        spacing: float,
        lattice: Lattice,
        offset: list[float] = (0, 0, 0),
        gamma_centered: bool = True,
    ):
        """
        Determines the k-point grid needed to achieve *at least* the provided k-point
        spacing (in Angstroms^-1). Spacing is defined as the minimum distance between
        any two k-points. The [a, b, c] grid rounds each value up to the nearest
        integer.

        Results in mode="automatic"
        """
        vectors = lattice.reciprocal_lattice.lengths

        # Calculate the number of k-points in each dimension
        k_points = [math.ceil(v / spacing) for v in vectors]

        # Ensuring an odd number of k-points in Monkhorst-Pack grids helps center
        # one k-point at the Brillouin zone's origin (Γ-point). This is achieved
        # by rounding down to the nearest even number, doubling it, and adding 1,
        # ensuring a well-defined center for accurate Brillouin zone sampling.
        # The convention aids in precise calculations of electronic states near
        # the zone center, contributing to the accuracy of electronic
        # structure calculations.
        if gamma_centered:
            k_points = [(pt // 2) * 2 + 1 for pt in k_points]
        # BUG: I'm not 100% sure this ensures a gamma centered point...
        # ChatGPT gave this suggestion and I'm just rolling with it.
        # This formula often results in a more dense grid than what was requested
        # so I'm sure an offset or something would be better

        return cls(
            mode="automatic",
            grid=k_points,
            offset=offset,
            weights=None,
        )

    @classmethod
    def from_density(
        cls,
        density: float,
        lattice: Lattice,
        **kwargs,
    ):
        """
        Determines the k-point grid needed to achieve *at least* the provided k-point
        density (in Angstroms^-3). The [a, b, c] grid rounds each value up to the
        nearest integer

        Results in mode="automatic"
        """
        spacing = 1 / (density ** (1 / 3))
        return cls.from_spacing(spacing=spacing, lattice=lattice, **kwargs)

    @classmethod
    def from_dynamic(
        cls,
        k_points: any,
        structure: Structure = None,
        lattice: Lattice = None,
    ):
        # TODO: check incorrect configs such as those that have BOTH
        # spacing and denisty. Or provide these with a grid. Or missing a lattice

        if structure:
            lattice = structure.lattice

        if isinstance(k_points, cls):
            return k_points

        elif isinstance(k_points, dict):
            if lattice and "spacing" in k_points:
                return cls.from_spacing(
                    spacing=k_points["spacing"],
                    lattice=lattice,
                )
            elif lattice and "density" in k_points:
                return cls.from_density(
                    density=k_points["density"],
                    lattice=lattice,
                )
            else:
                return cls(**k_points)

        return {}



import logging
import os
from functools import cached_property
from pathlib import Path

import yaml

import simmate
from simmate import website  # needed to specify location of built-in apps
from simmate.utilities import (
    deep_update,
    dotdict,
    get_conda_env,
    get_directory,
    str_to_datatype,
)


class SimmateSettings:
    """
    Configures Simmate settings
    """

    # TODO: consider using pydantic instead. I don't do this yet because of some
    # default values are dynamic and co-dependent, but I should revisit in the
    # future:
    #   https://github.com/pydantic/pydantic/issues/866

    # -------------------------------------------------------------------------

    @cached_property
    def final_settings(self) -> dict:
        """
        The final settings built from user-supplied settings and all defaults
        """
        settings = deep_update(
            default_dict=self.default_settings.copy(),
            override_dict=self.user_settings.copy(),
        )

        # clean variables
        # TODO: handle database.url input
        # TODO: use_docker inputs
        # CSRF_TRUSTED_ORIGINS = .split(",")
        # ALLOWED_HOSTS = .split(",")

        # REQUIRE_LOGIN = os.getenv("REQUIRE_LOGIN", "False") == "True"
        # # when setting REQUIRE_INTERNAL_LOGIN, set it to the allauth provider type
        # # (such as "microsoft")
        # REQUIRE_LOGIN_INTERNAL = os.getenv("REQUIRE_LOGIN_INTERNAL", "False")
        # if REQUIRE_LOGIN_INTERNAL == "False":
        #     REQUIRE_LOGIN_INTERNAL = False
        # else:
        #     assert REQUIRE_LOGIN_INTERNAL in ["microsoft", "google"]
        #     REQUIRE_LOGIN = True
        # # example: r'/apps/spotfire(.*)$'
        # REQUIRE_LOGIN_EXCEPTIONS = [
        #     e for e in os.getenv("REQUIRE_LOGIN_EXCEPTIONS", "").split(";") if e
        # ]
        # LOGIN_MESSAGE = os.getenv("LOGIN_MESSAGE", "")

        # Run compatibility checks (e.g. use_docker requires a 'docker run' cmd)
        # TODO

        return settings

    def __getattr__(self, name: str):
        """
        Makes all settings accessible as properties.

        This handles how `getattr(self, "example")` or `self.example` behaves
        when `example` is not actually defined.
        """
        if name not in self.final_settings.keys():
            raise Exception(f"Unknown property or setting: {name}")

        setting = self.final_settings.get(name, None)

        # if the property accessed is a dictionary, then we make it a dotdict
        # so that we can perform recursive dot access
        if isinstance(setting, dict):
            return dotdict(setting)
        else:
            return setting

    def show_settings(self, user_only: bool = False):
        """
        Takes the final simmate settings and prints them in a yaml format that is
        easier to read.
        """
        settings_to_print = self.final_settings if not user_only else self.user_settings
        print(yaml.dump(settings_to_print))

    def write_settings(self, filename: Path = None, settings: dict = "final"):
        """
        Writes simmate settings to a yaml file.
        Settings can be a dictionary, "final", or "user".
        """
        if not filename:
            filename = (
                settings.config_directory / f"_{settings.conda_env}-settings.yaml"
            )

        if settings == "final":
            settings = self.final_settings
        elif settings == "user":
            settings = self.user_settings
        else:
            assert isinstance(settings, dict)  # bug check

        logging.info(f"Writing settings to: '{filename}'")
        with filename.open("w") as file:
            content = yaml.dump(settings)
            file.write(content)

    def write_updated_settings(self, updates: dict):
        """
        For the convenience of beginners, some `simmate config` commands
        are to update specific settings. This writes updates to the currently
        used settings file
        """
        # TODO: Do I want to make a backup of the old settings...?

        # updates are only allowed if we have yaml config. This is bc we dont
        # want to set environment variables perminantly via Simmate.
        if self.settings_source is None:
            # default to most-specific config file
            source = self.config_directory / f"{self.conda_env}-settings.yaml"
        elif isinstance(self.settings_source, Path):
            source = self.settings_source
        else:
            raise Exception(
                "Updating your Simmate settings is only allowed when using"
                f"a YAML configuration. You are using {self.settings_source}."
            )

        logging.info("Updating the following settings:\n")
        print(yaml.dump(updates))

        final_user_settings = deep_update(self.user_settings, updates)
        self.write_settings(
            filename=source,
            settings=final_user_settings,
        )

    def add_apps_and_update(self, apps: list):
        """
        Given a list of app names (via python path to Config), it adds these
        the currently registered apps. This utility ensures no duplicates are
        added, and that default settings are not lost.
        """
        current_apps = self.final_settings["apps"]
        final_apps = current_apps.copy()
        for app in apps:
            if app not in final_apps:
                final_apps.append(app)
        # BUG: Is order of the apps important for some...?
        self.write_updated_settings(updates={"apps": final_apps})

    # -------------------------------------------------------------------------

    @cached_property
    def default_settings(self) -> dict:
        """
        The default settings for Simmate.

        Note that some of these settings (such as the database name) are
        determined dynamically, so default may vary across different
        simmate installations.

        Further, some settings update the default for others. An example of this
        is `use_docker` for Quantum Espresso, which will also change the
        `default_command` for this app.
        """
        return {
            "version": simmate.__version__,
            "apps": [
                "simmate.workflows.configs.BaseWorkflowsConfig",  # TODO - deprec
                "simmate.apps.configs.QuantumEspressoConfig",
                "simmate.apps.configs.VaspConfig",
                "simmate.apps.configs.BaderConfig",
                "simmate.apps.configs.EvolutionConfig",
                "simmate.apps.configs.MaterialsProjectConfig",
                # These apps may become defaults in the future:
                # "simmate.apps.configs.BadelfConfig",
                # "simmate.apps.configs.CleaseConfig",
                # "simmate.apps.configs.WarrenLabConfig",
            ],
            "database": self._default_database,
            "website": {
                # Sometimes we lock down the website to registered/approved users.
                # By default, we allow anonymous users to explore because this makes things like
                # REST API calls much easier for them. In special cases, such as industry, we
                # ONLY let users sign in via a specific allauth endpoint. An example of this
                # is Corteva limiting users to those approved via their Microsoft auth.
                "require_login": False,
                "require_login_internal": False,
                "require_login_exceptions": [],
                "login_message": None,
                # These allow server maintainers to override the homepage and profile views, which
                # is important if they involve loading custom apps/models for their templates.
                "home_view": None,
                "profile_view": None,
                "data": [
                    "simmate.database.third_parties.AflowPrototype",
                    # "simmate.database.third_parties.AflowStructure",  # Not allowed yet
                    "simmate.database.third_parties.CodStructure",
                    "simmate.database.third_parties.JarvisStructure",
                    "simmate.database.third_parties.MatprojStructure",
                    "simmate.database.third_parties.OqmdStructure",
                ],
                "social_oauth": {
                    "google": {"client_id": None, "secret": None},
                    "microsoft": {"client_id": None, "secret": None},
                    "github": {"client_id": None, "secret": None},
                },
                # django extras
                "debug": False,
                "allowed_hosts": ["127.0.0.1", "localhost"],
                # BUG-FIX: Django-unicorn ajax requests sometimes come from the server-side
                # ingress (url for k8s) or a nginx load balancer. To get past a 403 forbidden
                # result, we need to sometimes specify allowed origins for csrf.
                "csrf_trusted_origins": ["http://localhost"],
                # Keep the secret key used in production secret!
                # !!! I removed get_random_secret_key() so I don't have to sign out every time
                # while testing my server. I may change this back in the future.
                # from django.core.management.utils import get_random_secret_key
                "secret_key": "pocj6cunub4zi31r02vr5*5a2c(+_a0+(zsswa7fmus^o78v)r",
                # Settings for sending automated emails.
                # For example, this can be set up for GMail by...
                #   1. enabling IMAP (in gmail settings)
                #   2. Having 2-factor auth turned on
                #   3. Adding an App Password (in account settings)
                "email": {
                    "backend": "django.core.mail.backends.smtp.EmailBackend",  # this is the default
                    "host": "smtp.gmail.com",  # or outlook.office365.com
                    "port": 587,
                    "use_tls": False,
                    "host_user": "",
                    "host_password": "",
                    "from_email": "simmate.team@gmail.com",
                    "timeout": 5,
                    "subject_prefix": "[Simmate] ",
                    "account_verification": "none",  # when creating new accounts
                },
                # These people get an email when DEBUG=False
                "admins": [
                    ("jacksund", "jacksundberg123@gmail.com"),
                    ("jacksund-corteva", "jack.sundberg@corteva.com"),
                ],
            },
            # app-specific configs
            # TODO: consider moving these to the respective apps
            "bader": {
                "default_command": "bader CHGCAR",
                "docker": {
                    "enable": False,
                    "image": f"jacksund/bader:{simmate.__version__}",
                },
            },
            "vasp": {
                "default_command": "vasp_std > vasp.out",
                "parallelization": {
                    "ncore": None,
                    "kpar": None,
                },
                "docker": {
                    "enable": False,
                    "image": None,  # commercial license required
                },
            },
            "quantum_espresso": {
                "default_command": "pw.x < pwscf.in > pw-scf.out",
                "psuedo_dir": self.config_directory / "quantum_espresso" / "potentials",
                "docker": {
                    "enable": False,
                    "image": f"jacksund/quantum-espresso:{simmate.__version__}",
                },
            },
        }

    @cached_property
    def _default_database(self) -> Path:
        # if the user is in the (base) env or not using conda, then we will have a
        # value of "-database.sqlite3", which why we need strip() here.
        db_filename = (
            self.config_directory / f"{self.conda_env}-database.sqlite3".strip("-")
        )
        return {
            "engine": "django.db.backends.sqlite3",
            "name": db_filename,
        }

    @cached_property
    def database_backend(self) -> str:  # was... DATABASE_BACKEND
        if self.database.engine == "django.db.backends.sqlite3":
            return "sqlite3"
        elif self.database.engine == "django.db.backends.postgresql":
            return "postgresql"
        else:
            return "unknown"

    # -------------------------------------------------------------------------

    @cached_property
    def user_settings(self) -> dict:
        """
        A dictionary of keys-values that the user has set via the `setting_source`
        """
        source = self.settings_source

        # no user settings
        if not source:
            return {}
        # a yaml file
        elif isinstance(source, Path):
            with source.open() as file:
                return yaml.full_load(file)
        # environment variables
        elif source == "environment variables":
            return self._get_env_settings()

    @cached_property
    def settings_source(self) -> str | Path:
        """
        Where the settings are being loading from.

        There are several options for configuring settings that we check for, in order
        of priority:

        1. environment variables
        2. settings.yaml
        3. {conda env}-settings.yaml
        4. nothing (use the default)

        We do NOT allow for a mixture of these. This because tracking down where
        a setting came from would be tricky if there are both environment variables
        and settings files being used.
        """

        # 1. check if using environment variables
        prefix = "SIMMATE__"
        env_vars = dict(os.environ)
        if any(var.startswith(prefix) for var in env_vars.keys()):
            return "environment variables"

        # 2. settings.yaml file
        settings_file = self.config_directory / "settings.yaml"
        if settings_file.exists():
            return settings_file

        # 3. my_env-settings.yaml file
        settings_file = self.config_directory / f"{self.conda_env}-settings.yaml"
        if settings_file.exists():
            return settings_file

        # 4. Using 100% defaults
        return None

    def _get_env_settings(self) -> dict:
        """
        Parses all 'SIMMATE__' settings from available environment variables
        """

        # grab all variables associated with simmate and break down the variables
        # into a "normal" dictionary with proper python types
        # For example:
        #   {"SIMMATE__WEBSITE___REQUIRE_LOGIN": "true"}
        #   ...into..
        #   {"website": {"require_login": True}}
        user_settings = {}
        for key, value in dict(os.environ).items():
            # skip non-simmate settings
            if not key.startswith("SIMMATE__"):
                continue

            # skip simmate settings that are set via a helm chart or docker
            # container, but not updated by user. All of these cloud deployments
            # use "__skip__" as the default value
            if value == "__skip__":
                continue

            # Convert the value to the proper python type.
            value_cleaned = str_to_datatype(
                parameter=key,
                value=value,
                type_mappings=self._input_mappings,
            )
            # OPTIMIZE: There might be better ways to do this, such as
            # inspecting the attribute. Example with 'conda_env':
            #   SimmateSettings.conda_env.func.__annotations__["return"]

            # split the variable name into basic keys. Note, how we
            # throw out the first key because it is always "simmate".
            # We also convert to lowercase
            components = [c.lower() for c in key.split("__")[1:]]

            # reformat dictionary
            cleaned_setting = {components[-1]: value_cleaned}
            for sub_key in reversed(components[:-1]):
                cleaned_setting = {sub_key: cleaned_setting}

            # add to final set
            user_settings = deep_update(user_settings, cleaned_setting)

        return user_settings

    # -------------------------------------------------------------------------

    # Fixed settings that are automatically set

    @cached_property
    def conda_env(self) -> str:
        """
        Name of the conda environment being used.

        Some settings depend on the conda env name. This makes switching
        between different databases and settings as easy as activating different
        conda environments
        """
        return get_conda_env()

    @cached_property
    def config_directory(self) -> Path:
        """
        We check for all settings in the user's home directory and in a
        folder named "~/simmate/".
        For windows, this would be something like...
          C:\\Users\\exampleuser\\simmate\\extra_applications
        """
        # we use `get_directory` in order to create the folder if it does not exist.
        return get_directory(Path.home() / "simmate")

    @cached_property
    def django_directory(self) -> str:
        """
        Location of the `simmate.website` module & django "base"

        This directory is where simmate.website is located and helps us indicate
        where things like our templates or static files are located. We find this
        by looking at the import path to see where python installed it.
        """
        return Path(website.__file__).absolute().parent

    # -------------------------------------------------------------------------

    def _parse_input(self, input_str) -> str:
        """
        Converts something like
          "bader.docker.enable=True"
        into
          {"bader": {"docker": {"enable": True}}}
        """
        keys, value = input_str.split("=")
        keys = keys.split(".")

        final_dict = {}
        current_level = final_dict
        for key in keys[:-1]:
            current_level = current_level.setdefault(key, {})
        last_key = keys[-1]

        # convert value to the proper data type
        env_name = "SIMMATE__" + input_str.split("=")[0].replace(".", "__").upper()
        value_cleaned = str_to_datatype(
            parameter=env_name,
            value=value,
            type_mappings=self._input_mappings,
        )
        # OPTIMIZE: There might be better ways to do this, such as
        # inspecting the attribute. Example with 'conda_env':
        #   SimmateSettings.conda_env.func.__annotations__["return"]

        current_level[last_key] = value_cleaned

        return final_dict

    # OPTIMIZE: this is only used in the _get_env_settings method and should
    # depreciated in favor of type-inspecting
    _input_mappings: dict = {
        "SIMMATE__APPS": list[str],
        "SIMMATE__DATABASE__ENGINE": str,
        "SIMMATE__DATABASE__HOST": str,
        "SIMMATE__DATABASE__NAME": str,
        "SIMMATE__DATABASE__USER": str,
        "SIMMATE__DATABASE__PASSWORD": str,
        "SIMMATE__DATABASE__PORT": int,
        "SIMMATE__WEBSITE__ALLOWED_HOSTS": list[str],
        "SIMMATE__WEBSITE__CSRF_TRUSTED_ORIGINS": list[str],
        "SIMMATE__WEBSITE__DATA": list[str],
        "SIMMATE__WEBSITE__DEBUG": bool,
        "SIMMATE__WEBSITE__EMAILS__FROM_EMAIL": str,
        "SIMMATE__WEBSITE__EMAILS__HOST": str,
        "SIMMATE__WEBSITE__EMAILS__PORT": int,
        "SIMMATE__WEBSITE__EMAILS__TIMEOUT": int,  # or float...?
        "SIMMATE__WEBSITE__HOME_VIEW": str,
        "SIMMATE__WEBSITE__LOGIN_MESSAGE": str,
        "SIMMATE__WEBSITE__PROFILE_VIEW": str,
        "SIMMATE__WEBSITE__REQUIRE_LOGIN": bool,
        "SIMMATE__WEBSITE__REQUIRE_LOGIN_EXCEPTIONS": list[str],
        "SIMMATE__WEBSITE__REQUIRE_LOGIN_INTERNAL": bool,
        "SIMMATE__WEBSITE__SOCIAL_OAUTH__MICROSOFT__CLIENT_ID": str,
        "SIMMATE__WEBSITE__SOCIAL_OAUTH__MICROSOFT__SECRET": str,
        "SIMMATE__BADER__DOCKER__ENABLE": bool,
        "SIMMATE__QUANTUM_ESPRESSO__DOCKER__ENABLE": bool,
        "SIMMATE__VASP__DOCKER__ENABLE": bool,
    }


# initialize all settings
settings = SimmateSettings()



# -*- coding: utf-8 -*-

from simmate.apps.warren_lab.workflows.relaxation.pbe import (
    Relaxation__Vasp__WarrenLabPbe,
)


class Relaxation__Vasp__WarrenLabScan(Relaxation__Vasp__WarrenLabPbe):
    """
    Runs a VASP relaxation calculation using Warren Lab SCAN functional settings.
    """

    description_doc_short = "Warren Lab presets for SCAN geometry optimization"

    _incar_updates = dict(
        ALGO="All",
        EDIFF=1.0e-05,
        EDIFFG=-0.02,
        ENAUG=1360,
        ENCUT=680,
        LMIXTAU=True,
        METAGGA="R2scan",
        KSPACING=0.22,
        ISMEAR=2,
        SIGMA=0.2,
        IVDW=0,  # No d3 coefficients for SCAN so we just turn of VDW corrections
    )



# -*- coding: utf-8 -*-

import logging
from pathlib import Path

from simmate.apps.vasp.inputs import Incar, Poscar, Potcar
from simmate.apps.vasp.workflows.base import VaspWorkflow
from simmate.toolkit.diffusion import MigrationImages


class VaspNebFromImagesWorkflow(VaspWorkflow):
    """
    A base class for Nudged Elastic Band (NEB) calculations. This is not meant
    to be used directly but instead should be inherited from.

    Note, NEB tasks are very different from all other VASP tasks!

    The first big difference is that it takes a list of structures instead of
    just one structure. This means that instead of "structure=...", you should
    actually do "structures=[structure1, structure2, structure3, ...]", where
    the list of structures is your list of images.

    The second big difference is that VASP uses a different folder setup when
    running these calculations. It has a series of folders named 00, 01, 02, ... N,
    where 00 is the starting image, N is the endpoint image, and 01 to (N-1) are
    the midpoint images. Simmate handles this inside the task, but knowing this
    may be useful if you'd like to make your own variation of this class.
    """

    _parameter_methods = VaspWorkflow._parameter_methods.copy()
    _parameter_methods.remove("_get_clean_structure")

    # NEB does not require a POSCAR file because input structures are organized
    # into folders.
    required_files = ["INCAR", "POTCAR"]

    description_doc_short = "runs NEB using a series of structures images as input"

    @classmethod
    def setup(
        cls,
        directory: Path,
        migration_images: MigrationImages,
        **kwargs,
    ):
        """
        Writes input files for a NEB calculation. Each structure image recieves
        it's own folder within the parent directory.
        """

        # run some prechecks to make sure the user has everything set up properly.

        # One common mistake is to mislabel the number of images in the
        # INCAR file.
        # first, we check if the user set this.
        nimages = cls.incar.get("IMAGES")
        if nimages:
            # if so, we check that it was set correctly. It should be equal to
            # the number of structures minus 2 (because we don't count the
            # start and end images here.)
            if nimages != (len(migration_images) - 2):
                raise Exception(
                    "IMAGES looks to be improperly set! This value should not"
                    " include the start/end images -- so make sure you counted"
                    " properly. Alternatively, you also can remove this keyword"
                    " from your INCAR and Simmate will provide it automatically"
                    " for you."
                )

        # TODO: add a precheck that ensures the number of cores VASP is ran on
        # is also divisible by the number of images. For example...
        # "mpirun -n 16 vasp" will not work for IMAGES=3 because 16 is not
        # divisible by 3. But this also may be better suited for an ErrorHandler.
        # An example error message from from VASP is...
        #       "M_divide: can not subdivide 16 nodes by 3"

        # Here, each image (start to end structures) is put inside of its own
        # folder. We make those folders here, where they are named 00, 01, 02...N
        # Also recall that "structure" is really a list of structures here.
        for i, image in enumerate(migration_images):
            # first make establish the foldername
            # The zfill function converts numbers from "1" to "01" for us
            foldername = directory / str(i).zfill(2)
            # see if the folder exists, and if not, make it
            if not foldername.exists():
                foldername.mkdir()
            # now write the poscar file inside the folder
            Poscar.to_file(image, foldername / "POSCAR")

        # We also need to check if the user set IMAGES in the INCAR. If not,
        # we set that for them here. Note, we use the "pop" method which removes
        # the IMAGES__auto while grabbing its value. Because we are modifying
        # the incar dictionary here, we must make a copy of it -- this ensures
        # no bugs when this task is called in parallel.
        incar = cls.incar.copy()
        # !!! Should this code be moved to the INCAR class? Or would that require
        # too much reworking to allow INCAR to accept a list of structures?
        if not incar.get("IMAGES") and incar.pop("IMAGES__auto", None):
            incar["IMAGES"] = len(migration_images) - 2

        # Combine our base incar settings with those of our parallel settings
        # and then write the incar file
        incar = Incar(**incar)
        incar.to_file(
            filename=directory / "INCAR",
            # we can use the start image for our structure -- as all structures
            # should give the same result.
            structure=migration_images[0],
        )

        # if KSPACING is not provided in the incar AND kpoints is attached to this
        # class instance, then we write the KPOINTS file
        if cls.kpoints and ("KSPACING" not in cls.incar):
            raise Exception(
                "Custom KPOINTS are not supported by Simmate yet. "
                "Please use KSPACING in your INCAR instead."
            )

        # write the POTCAR file in this folder as well. Only one is needed and
        # all images will use refer to this POTCAR in the base directory
        Potcar.to_file_from_type(
            # we can use the start image for our structure -- as all structures
            # should give the same result.
            migration_images[0].composition.elements,
            cls.functional,
            directory / "POTCAR",
            cls.potcar_mappings,
        )

        # For the user's reference, we also like to write an image of the
        # starting path to a cif file. This can be slow for large structures
        # (>1s), but it is very little time compared to a full NEB run.
        path_vis = migration_images.get_sum_structure()
        path_vis.to(
            filename=str(directory / "simmate_path_relaxed_idpp.cif"),
            fmt="cif",
        )

    @classmethod
    def setup_restart(cls, directory: Path, **kwargs):
        """
        From a working directory of a past calculation, sets up for the calculation
        to be restarted.
        """
        logging.warning("CONTCARs are not yet copied to POSCARs for NEB restarts.")

        # establish filenames
        stopcar_filename = directory / "STOPCAR"

        # delete the stopcar if it exists
        if stopcar_filename.exists():
            stopcar_filename.unlink()



# -*- coding: utf-8 -*-

import warnings
from pathlib import Path

import pandas
from pymatgen.io.vasp import Potcar
from pymatgen.io.vasp.outputs import Chgcar


def ACF(directory: Path = None, filename="ACF.dat"):
    # grab working directory if one wasn't provided
    if not directory:
        directory = Path.cwd()

    # convert to path obj
    filename = directory / filename

    # open the file, grab the lines, and then close it
    with filename.open() as file:
        lines = file.readlines()

    # establish the headers. Note that I ignore the '#' column as this is
    # just site index.
    headers = ("x", "y", "z", "charge", "min_dist", "atomic_vol")

    # create a list that we will load data into
    bader_data = []
    # The first 2 lines are header and the final 4 lines are the footer. This is always
    # true so we don't need to iterate through those. The data we want is between the
    # header and footer so that's what we loop through.
    for line in lines[2:-4]:
        # By running strip, we convert the line from a string to a list of
        # The values are all still strings, so we convert them to int/floats
        # before saving. I add [1:] because the first value is just '#' which
        # is site index and we dont need.
        line_data = [eval(value) for value in line.split()[1:]]
        # add the line data to our ouput
        bader_data.append(line_data)

    # convert the list to a pandas dataframe
    dataframe = pandas.DataFrame(
        data=bader_data,
        columns=headers,
    )

    # Extra data is included in the footer that we can grab too. For each line, the data
    # is a float that is at the end of the line, hence the split()[-1].
    extra_data = {
        "vacuum_charge": float(lines[-3].split()[-1]),
        "vacuum_volume": float(lines[-2].split()[-1]),
        "nelectrons": float(lines[-1].split()[-1]),
    }

    # The remaining code is to analyze the results and calculate extra
    # information such as the final oxidation states. This requires extra
    # files to be present, such as from a vasp calculation

    potcar_filename = directory / "POTCAR"
    chgcar_filename = directory / "CHGCAR"
    CHGCAR_w_empty_atoms_filename = directory / "CHGCAR_w_empty_atoms"  # SPECIAL CASE

    # check if the required vasp files are present before doing the workup
    if potcar_filename.exists() and (
        chgcar_filename.exists() or CHGCAR_w_empty_atoms_filename.exists()
    ):
        # load the electron counts used by VASP from the POTCAR files
        # OPTIMIZE: this can be much faster if I have a reference file
        # BUG: pymatgen prints warnings if potcars don't match. we ignore these
        with warnings.catch_warnings(record=True):
            potcars = Potcar.from_file(potcar_filename)
        nelectron_data = {}
        # the result is a list because there can be multiple element potcars
        # in the file (e.g. for NaCl, POTCAR = POTCAR_Na + POTCAR_Cl)
        for potcar in potcars:
            nelectron_data.update({potcar.element: potcar.nelectrons})

        # SPECIAL CASE: in scenarios where empty atoms are added to the structure,
        # we should grab that modified structure instead of the one from the POSCAR.
        # the empty file will always take preference
        if CHGCAR_w_empty_atoms_filename.exists():
            chgcar = Chgcar.from_file(CHGCAR_w_empty_atoms_filename)
            structure = chgcar.structure
            # We typically use hydrogen ("H") as the empty atom, so we will
            # need to add this to our element list for oxidation analysis.
            # We use 0 for electron count because this is an 'empty' atom, and
            # not actually Hydrogen.
            # Note, this isn't ALWAYS H, so we iterate through the structure
            # and whatever element is missing from our nelectron_data is said
            # to be an empty and we set to 0.
            for element in structure.composition:
                if element.symbol not in nelectron_data.keys():
                    nelectron_data.update({element.symbol: 0})

        # otherwise, grab the structure from the CHGCAR
        # OPTIMIZE: consider grabbing from the POSCAR or CONTCAR for speed
        else:
            chgcar = Chgcar.from_file(chgcar_filename)
            structure = chgcar.structure

        # Calculate the oxidation state of each site where it is simply the
        # change in number of electrons associated with it from vasp potcar vs
        # the bader charge I also add the element strings for filtering functionality
        elements = []
        oxi_state_data = []
        for site, site_charge in zip(structure, dataframe.charge.values):
            element_str = site.specie.name
            elements.append(element_str)
            oxi_state = nelectron_data[element_str] - site_charge
            oxi_state_data.append(oxi_state)

        # add the new column to the dataframe
        dataframe = dataframe.assign(
            oxidation_state=oxi_state_data,
            element=elements,
        )
        # !!! There are multiple ways to do this, but I don't know which is best
        # dataframe["oxidation_state"] = pandas.Series(
        #     oxi_state_data, index=dataframe.index)

    return dataframe, extra_data



# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .matproj_full import ElectronicStructure__Vasp__MatprojFull
from .matproj_hse_full import ElectronicStructure__Vasp__MatprojHseFull
from .matproj_band_structure import ElectronicStructure__Vasp__MatprojBandStructure
from .matproj_density_of_states import ElectronicStructure__Vasp__MatprojDensityOfStates
from .matproj_band_structure_hse import (
    ElectronicStructure__Vasp__MatprojBandStructureHse,
)
from .matproj_density_of_states_hse import (
    ElectronicStructure__Vasp__MatprojDensityOfStatesHse,
)



# -*- coding: utf-8 -*-

import pandas

from simmate.apps.evolution.selectors import Selector


class TruncatedSelection(Selector):
    """
    Truncated selection limits the parent selection to the top X% of individuals,
    and then within this top section, every individual has an equal chance of
    being selected.
    """

    @staticmethod
    def select(
        nselect: int,
        individuals: pandas.DataFrame,
        fitness_column: str,
        percentile: float = 0.05,
        # at min/max, make sure we can select from this number of individuals
        ntruncate_min: int = 5,
        ntruncate_max: int = 50,
        # whether we can select the same individual more than once
        allow_duplicate: bool = True,
    ) -> pandas.DataFrame:
        # truncate the population to those with energies in the lowest X%. This
        # value should be greater or equal to our minimum set above
        ntruncate = int(len(individuals) * percentile)
        if ntruncate < ntruncate_min:
            ntruncate = ntruncate_min
        if ntruncate > ntruncate_max:
            ntruncate = ntruncate_max

        # limit our potential selection to the top X% and then randomly select
        # the correct number of parents
        df_truncated = individuals.nsmallest(ntruncate, fitness_column)
        df_parents = df_truncated.sample(nselect, replace=allow_duplicate)

        # return the list of indexes to be selected
        return df_parents



# -*- coding: utf-8 -*-

import logging

import numpy
import scipy
from django.core.exceptions import MultipleObjectsReturned
from django.utils import timezone
from rich.progress import track

from simmate.toolkit import Structure
from simmate.toolkit.validators import Validator
from simmate.utilities import chunk_list


class FingerprintValidator(Validator):
    comparison_mode: str = "linalg_norm"
    """
    How fingerprints distances should be determined. Options are:
        - linalg_norm (euclidean distance)
        - cos (cosine similarity)
        - custom

    For 'custom', you must write a custom 'get_fingerprint_distance(fp1, fp2)'
    static method that will be used.
    """

    distance_tolerance: float = 0.005
    """
    The value used to signify different structures
    """

    def __init__(
        self,
        distance_tolerance: float = None,  # defaults to class attr
        structure_pool: list[Structure] = [],  # OR a queryset from a Structure table
        use_database: bool = False,
        **kwargs,
    ):
        self.use_database = use_database
        self.distance_tolerance = distance_tolerance or self.distance_tolerance

        # setup featurizer with the given composition
        self.featurizer = self.get_featurizer(**kwargs)

        # see if we are working with a list of toolkit structures or a queryset
        if isinstance(structure_pool, list):
            # When the database isn't being used, we still need set these
            # variable as none to help with some other methods.
            self.structure_pool_queryset = "local_only"
        else:
            # we store the queryset as an attribute because we may want to
            # update the structure pool later on.
            self.structure_pool_queryset = structure_pool

        # If we are using the database, that means we are storing fingerprints
        # inside a cloud database table. Here, we set up the parent table entry.
        if use_database:
            if self.structure_pool_queryset == "local_only":
                raise Exception(
                    "You can only use database features if a queryset is given "
                    "for your structure_pool input."
                )

            # The fingerprint database table can store many different
            # fingerprints -- both from different fingerprint featurizers AND
            # using different settings for a given featurizer. We therefore
            # need to store the init kwargs with each fingerprint in the database
            from simmate.database.base_data_types import FingerprintPool
            from simmate.engine import Workflow

            self.init_kwargs = Workflow._serialize_parameters(
                distance_tolerance=self.distance_tolerance,
                **kwargs,
            )
            # TODO: _serialize_parameters should be a utility and not
            # attached to the workflow class

            self.database_pool = FingerprintPool.objects.get_or_create(
                method=self.name,
                init_kwargs=self.init_kwargs,
                database_table=structure_pool.model.table_name,
            )[0]
            # BUG: There is a race condition here. If a pool is started up from
            # multiple locations, this could result in duplicate pools.

            # we also keep a log of the last update so we only grab new structures
            # each time we update the database. To start, we set this as the
            # earliest possible date, which tells our update_fingerprint_pool
            # method to include ALL structures
            self.last_update = timezone.make_aware(
                timezone.datetime.min, timezone.get_default_timezone()
            )

        # next we address what initial structures were given.

        # check if we were given a list of pymatgen structures. If so, we can
        # just set and store things locally.
        if isinstance(structure_pool, list):
            # If so, we generate the fingerprint for each of the initial input structures
            # We convert this to a numpy array for speed improvement at later stages
            fingerprints = numpy.array(
                [
                    self._get_fingerprint(structure)
                    for structure in track(structure_pool)
                ]
            )
            # OPTIMIZE: this can be slow and I should support parallel featurization.

            sources = [structure.source for structure in structure_pool]

            self._add_many_to_pool(fingerprints, sources)

        # otherwise we have a queryset that should be used to populate the
        # fingerprint database
        else:
            self.fingerprint_pool = numpy.array([])
            self.source_pool = []
            self.update_fingerprint_pool()

    # -------------------------------------------------------------------------
    # Methods that can be overwritten when creating a new subclass
    # -------------------------------------------------------------------------

    @staticmethod
    def get_featurizer(self, **kwargs):
        raise NotImplementedError(
            "make sure you add a custom 'get_featurizer' method. This should "
            "return a featurizer object to be used for generating fingerprints."
        )

    def get_fingerprint_distance(self, fingerprint1, fingerprint2) -> float:
        raise NotImplementedError(
            "If you set the class attribute 'comparison_mode' to 'custom', make "
            "sure you also add a custom 'get_fingerprint_distance' method. This "
            "should return a distance (float value) ."
        )

    @staticmethod
    def format_fingerprint(fingerprint):
        return fingerprint  # does nothing by default

    # -------------------------------------------------------------------------
    # Core methods that generate and compare fingerprints
    # -------------------------------------------------------------------------

    def check_structure(
        self,
        structure: Structure,
        add_unique_to_pool: bool = True,
    ):
        # TODO:
        # maybe add a check to see if a matching source is already in the
        # the source_pool.

        # make the fingerprint
        fingerprint = self._get_fingerprint(structure)

        # compare this new fingerprint to all others
        is_unique = self._check_fingerprint(fingerprint, self.fingerprint_pool)

        # add this new fingerprint to the database if it was requested.
        if is_unique and add_unique_to_pool:
            # BUG-FIX:
            # in case a pymatgen structure was given, set the source to {}
            if not hasattr(structure, "source"):
                structure.source = {}

            self._add_to_pool(fingerprint, structure.source)

        # Return that we were successful
        return is_unique

    def _check_fingerprint(
        self,
        fingerprint: numpy.array,
        fingerprint_pool: list[numpy.array],
    ):
        # We now want to get the distance of this fingerprint relative to all others.
        # If the distance is within the specified tolerance, then the structures
        # are too similar - and we return  for a failure.
        is_unique = True  # consider it unique until proven otherwise
        # min_dist = 9999  # start with high value
        for fingerprint2 in fingerprint_pool:
            # check fingerprint based on the mode set
            if self.comparison_mode == "linalg_norm":
                distance = numpy.linalg.norm(fingerprint - fingerprint2)
            elif self.comparison_mode == "cos":
                distance = scipy.spatial.distance.cosine(fingerprint, fingerprint2)
            elif self.comparison_mode == "custom":
                distance = self.get_fingerprint_distance(fingerprint, fingerprint2)
            else:
                raise NotImplementedError("Unknown comparison_mode provided.")

            # keep track of the most similar structure
            # if min_dist > distance:
            #     min_dist = distance

            # and determine if we have a match
            if distance < self.distance_tolerance:
                # we can end the whole for-loop as soon as one structure is
                # deemed too similar
                is_unique = False
                break

        # If we make it through all structures and no distance is below the
        # tolerance, then we have a new and unique fingerprint

        return is_unique

    def _get_fingerprint(self, structure: Structure):
        # make the fingerprint for this structure into a numpy array for speed
        fingerprint = numpy.array(self.featurizer.featurize(structure))

        # apply any extra formatting
        fingerprint = self.format_fingerprint(fingerprint)

        return fingerprint

    # -------------------------------------------------------------------------
    # Methods that populate the pool and database with information
    # -------------------------------------------------------------------------

    def update_fingerprint_pool(self):
        if not self.use_database or self.structure_pool_queryset == "local_only":
            raise Exception(
                "This method should only be used when your structure pool"
                " is based on a Simmate database table!"
            )

        # what if a structure was added to the table WHILE I was running the
        # last query? This sets up a race condition with new structures and my
        # last_update timestamp. To avoid this, I immediately update our
        # timestamp, rather than doing it after the query
        last_update_safe = self.last_update
        self.last_update = timezone.now()

        # Now grab all the new structure ids that need to be added to the database
        new_ids = self.structure_pool_queryset.filter(
            created_at__gte=last_update_safe
        ).values_list("id", flat=True)

        # If there aren't any new structures, just exit without printing the
        # message and progress bar
        if not new_ids:
            return

        # first we can check if these structures have fingerprints already
        # calculated and load those.
        if self.use_database:
            logging.info("Checking database for already-calculated fingerprints")

            # a single query might be too large for some querysets. Several
            # smaller queries are more stable so we never grab more than 500
            # fingerprints at a time.
            all_results = []

            for query_chunk in chunk_list(new_ids, chunk_size=1000):
                query = self.database_pool.fingerprints.filter(
                    database_id__in=query_chunk
                )

                # BUG: there is a race condition that sometimes adds duplicate
                # fingerprints to the database. We add distinct to keep our
                # query smaller and just grab one. Distinct is also not supported
                # by sqlite
                from simmate.configuration import settings

                if settings.database.engine != "django.db.backends.sqlite3":
                    query = query.distinct("database_id")

                all_results += list(query)

            # the query does not return the ids in the same order that new_ids
            # was given. Order is important when finding unique structures, so
            # we need to reorder the query results here
            all_data_dict = {entry.database_id: entry for entry in all_results}
            all_data_ordered = [
                all_data_dict[id] for id in new_ids if id in all_data_dict.keys()
            ]

            fingerprints = [entry.fingerprint for entry in all_data_ordered]
            sources = [entry.source for entry in all_data_ordered]
            self._add_many_to_pool(fingerprints, sources, skip_database=True)

            # reset the new_structures list to those that are actually still needed
            existing_ids = [fp.database_id for fp in all_results]
            new_ids = [i for i in new_ids if i not in existing_ids]

        # same as before -- exit if there aren't any new ids
        if not new_ids:
            return

        logging.info(f"Found {len(new_ids)} new structure(s) for the fingerprint pool.")

        new_structures = self.structure_pool_queryset.filter(
            id__in=new_ids
        ).to_toolkit()

        # OPTIMIZE: I attempted to calculate fingerprints with Dask, but without
        # any luck. Ends up crashing in many scenarios.
        # from simmate.configuration.dask import get_dask_client
        # with get_dask_client() as client:
        #     futures = [client.submit(self._get_fingerprint, s) for s in new_structures]
        #     fingerprints = [future.result() for future in track(futures)]

        # calculate each fingerprint and add it to the database
        fingerprints = [
            self._get_fingerprint(structure) for structure in track(new_structures)
        ]
        sources = [structure.source for structure in new_structures]
        self._add_many_to_pool(fingerprints, sources)

    def _add_many_to_pool(self, fingerprints, sources, skip_database: bool = False):
        """
        Efficiently adds many fingerprints to the pool.
        """

        # source
        self.source_pool += sources

        # fingerprint
        # Numpy arrays begin differently when empty
        if self.fingerprint_pool.size == 0:
            self.fingerprint_pool = numpy.array(fingerprints)
        else:
            self.fingerprint_pool = numpy.append(
                self.fingerprint_pool, fingerprints, axis=0
            )

        # store in database
        if not skip_database:
            for fingerprint, source in zip(fingerprints, sources):
                self._add_to_database(fingerprint, source)
            # OPTIMIZE: I'm not sure how to do this in a single query
            # https://stackoverflow.com/questions/27047630

    def _add_to_pool(
        self,
        fingerprint,
        source: dict = {},
        skip_database: bool = False,
    ):
        """
        Adds a new fingerprint to the pool.

        This is very slow for >1000 additions though. For bulk additions, use
        _add_many_to_pool method instead. See:
            https://stackoverflow.com/questions/7133885
        """

        # source
        self.source_pool.append(source)

        # fingerprint
        # Numpy arrays begin differently when empty
        if self.fingerprint_pool.size == 0:
            self.fingerprint_pool = numpy.array([fingerprint])
        else:
            self.fingerprint_pool = numpy.append(
                self.fingerprint_pool, [fingerprint], axis=0
            )

        # store in database
        if not skip_database:
            self._add_to_database(fingerprint, source)

    def _add_to_database(self, fingerprint, source):
        # as an extra, we save the result to our database so that this
        # fingerprint doesn't need to be calculated again
        if self.use_database and source.get("database_id"):
            try:
                self.database_pool.fingerprints.update_or_create(
                    database_id=source.get("database_id"),
                    defaults=dict(fingerprint=list(fingerprint)),
                )
            # BUG: there is a race condition that sometimes triggers multiple
            # fingerprints to be added. If this happens, we just move on. The
            # key thing is that the fingerprint is there.
            except MultipleObjectsReturned:
                # TODO: maybe delete the duplicates here...?
                pass

    # -------------------------------------------------------------------------
    # Extra high level methods that are useful for analyzing a structure pool
    # -------------------------------------------------------------------------

    def remove_duplicates(self, structures: list[Structure]) -> list[Structure]:
        raise NotImplementedError("This method is still under development")
        # This might need to be a class method because I don't want to use
        # an already existing fingerprint_pool
        structures_unique = []
        for structure in track(structures):
            if self.check_structure(structure):
                structures_unique.append(structure)

    def get_unique_from_pool(self) -> list[Structure]:
        # order of the input structures is important for this method
        if (
            self.structure_pool_queryset
            and not self.structure_pool_queryset.query.order_by
        ):
            logging.warning(
                "The order of your structures is important. You should set "
                "order_by('some_column') on your database search. If duplicates "
                " are found, only the first structure will be returned."
            )

        logging.info("Isolating unique structures")

        # The first structure in our pool is unique by default so this
        # starts out our list
        unique_sources = [self.source_pool[0]]
        unique_fingerprints = numpy.array([self.fingerprint_pool[0]])

        # The rest need to be checked one at a time
        for source, fingerprint in track(
            list(zip(self.source_pool, self.fingerprint_pool))
        ):
            is_unique = self._check_fingerprint(
                fingerprint,
                unique_fingerprints,
            )
            if is_unique:
                unique_sources.append(source)
                unique_fingerprints = numpy.append(
                    unique_fingerprints, [fingerprint], axis=0
                )

        logging.info(
            f"{len(unique_sources)} unique entries found. "
            "Pulling structures from database."
        )

        # collect the unique sources so that we can make a single query.
        from simmate.file_converters.structure.database import DatabaseAdapter

        structures = DatabaseAdapter.get_toolkits_from_database_dicts(unique_sources)

        return structures



# -*- coding: utf-8 -*-

import logging

from rich import print

import simmate.toolkit.creators as creation_module
import simmate.toolkit.transformations as transform_module
import simmate.toolkit.transformations.from_ase as ase_transform_module
from simmate.database.base_data_types import DatabaseTable, table_column
from simmate.engine.execution import WorkItem
from simmate.toolkit import Composition


class SteadystateSource(DatabaseTable):
    class Meta:
        app_label = "workflows"

    name = table_column.CharField(max_length=50)
    kwargs = table_column.JSONField(default=dict)

    nsteadystate_target = table_column.FloatField(null=True, blank=True)

    is_creator = table_column.BooleanField()
    is_transformation = table_column.BooleanField()

    # This list limits to ids that are submitted or running
    workitem_ids = table_column.JSONField(default=list)

    search = table_column.ForeignKey(
        "FixedCompositionSearch",
        on_delete=table_column.CASCADE,
        related_name="steadystate_sources",
    )

    def update_flow_workitem_ids(self):
        """
        Queries Prefect to see how many workflows are in a scheduled, running,
        or pending state from the list of run ids that are associate with this
        structure source.
        """

        # Before updating our run ids, we first want to see if any of them
        # failed (or "E" for ERRORED) -- so that we can warn the use to
        # check their logs.
        failed_ids = WorkItem.objects.filter(
            id__in=self.workitem_ids,
            status__in=["E", "C"],
        ).values_list("id", flat=True)

        if failed_ids:
            logging.warning(
                f"The following WorkItem IDs failed or were cancelled: {list(failed_ids)}."
            )

            print("\n------------------------------------------------------------\n")
            print("Make sure you check your worker logs for more info.\n")
            print("You can preview ALL failed jobs in the command line:")
            print("simmate engine show-error-summary\n")
            print("Or view a specific error (w. full trackback) in python:")
            print("from simmate.engine.execution import WorkItem")
            print(f"item = WorkItem.objects.get(id={failed_ids[0]})")
            print("item.result()\n")
            print(
                "The call to `result()` will raise the error that caused your job "
                "to fail. Please report this error if you think it's a bug "
                "with Simmate.\n "
            )
            print(
                "Note, ~0-1% of calculations failing can be normal during an "
                "evolutionary search because generated structures may be "
                "unreasonable (and cause programs like VASP to crash). "
                "However, >1% can indicate a serious problem.\n"
            )
            print("------------------------------------------------------------\n")

        # check which ids are still running. Note, this is a separate
        # method in case it's an async call to Prefect client
        still_running_ids = self._check_still_running_ids(self.workitem_ids)

        # we now have our new list of IDs! Let's update it to the database
        self.workitem_ids = still_running_ids
        self.save()

        return still_running_ids

    @property
    def nflow_runs(self):
        # update our ids before we report how many there are.
        runs = self.update_flow_workitem_ids()
        # now the currently running ones is just the length of ids!
        return len(runs)

    @staticmethod
    def _check_still_running_ids(workitem_ids):
        """
        Queries Prefect to see check on a list of flow run ids and determines
        which ones are still in a scheduled, running, or pending state.
        From the list of ids given, it will return a list of the ones that
        haven't finished yet.
        This is normally used within `update_flow_workitem_ids` and shouldn't
        be called directly.
        """
        # The reason we have this code as a separate method is because we want
        # to allow isolation of Prefect's async calls from Django's
        # sync-restricted calls (i.e. django raises errors if called
        # within an async context).
        still_running_ids = WorkItem.objects.filter(
            id__in=workitem_ids,
            status__in=["P", "R"],
        ).values_list("id", flat=True)
        return list(still_running_ids)

    def to_toolkit(self):
        if self.is_transformation:
            return self._init_transformation()
        elif self.is_creator:
            return self._init_creator()

    def _init_transformation(self):
        if not self.is_transformation:
            raise Exception("This should not be called on non-transformations")

        # Consider moving to _deserialize_parameters. Only issue is that
        # I can't serialize these classes yet.
        if self.name.startswith("from_ase."):
            # all start with "from_ase" so I assume that import for now
            ase_class_str = self.name.split(".")[-1]
            transformation_class = getattr(ase_transform_module, ase_class_str)
            transformer = transformation_class(**self.kwargs)
        # !!! There aren't any common transformations that don't accept composition
        # as an input, but I expect this to change in the future.
        elif self.name in ["ExtremeSymmetry"]:
            transformation_class = getattr(transform_module, self.name)
            transformer = transformation_class(**self.kwargs)
        else:
            raise Exception(f"Transformation class {self.name} could not be found.")
        return transformer

    def _init_creator(self):
        if not self.is_creator:
            raise Exception("This should not be called on non-creators")

        composition = Composition(self.search.composition)

        # Consider moving to _deserialize_parameters. Only issue is that
        # I can't serialize these classes yet.
        try:
            creator_class = getattr(creation_module, self.name)
            creator = creator_class(composition, **self.kwargs)
        except:
            raise Exception(f"Creator class {self.name} could not be found.")
        return creator

    # This method is for Prefect 2.0
    # from simmate.utilities import async_to_sync
    # @staticmethod
    # @async_to_sync
    # async def _check_still_running_ids(workitem_ids):
    #     """
    #     Queries Prefect to see check on a list of flow run ids and determines
    #     which ones are still in a scheduled, running, or pending state.
    #     From the list of ids given, it will return a list of the ones that
    #     haven't finished yet.
    #     This is normally used within `update_flow_workitem_ids` and shouldn't
    #     be called directly.
    #     """
    #     raise NotImplementedError("porting to general executor")
    #     from prefect.client import get_client
    #     from prefect.orion.schemas.filters import FlowRunFilter
    #     # The reason we have this code as a separate method is because we want
    #     # to isolate Prefect's async calls from Django's sync-restricted calls
    #     # (i.e. django raises errors if called within an async context).
    #     async with get_client() as client:
    #         response = await client.read_flow_runs(
    #             flow_run_filter=FlowRunFilter(
    #                 id={"any_": workitem_ids},
    #                 state={"type": {"any_": ["SCHEDULED", "PENDING", "RUNNING"]}},
    #             ),
    #         )
    #     still_running_ids = [str(entry.id) for entry in response]
    #     return still_running_ids



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.static_energy.matproj import (
    StaticEnergy__Vasp__Matproj,
)


class Nmr__Vasp__MatprojChemicalShifts(StaticEnergy__Vasp__Matproj):
    """
    This task is a reimplementation of pymatgen's
    [MPNMRSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPNonSCFSet)
    with mode="cs" (Chemical Shift).
    """

    _incar_updates = dict(
        LCHIMAG=True,
        EDIFF=-1.0e-10,
        ISYM=0,
        LCHARG=False,
        LNMR_SYM_RED=True,
        NELMIN=10,
        NSLPLINE=True,
        PREC="ACCURATE",
        SIGMA=0.01,
    )



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.matproj import (
    Relaxation__Vasp__Matproj,
)


class Relaxation__Vasp__MvlGrainboundary(Relaxation__Vasp__Matproj):
    """
    This task is a reimplementation of pymatgen's
    [MVLGBSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MVLGBSet).
    """

    description_doc_short = "based on pymatgen's MVLGBSet"

    _incar_updates = dict(
        LCHARG=False,
        NELM=60,
        PREC="Normal",
        EDIFFG=-0.02,
        ICHARG=0,
        NSW=200,
        EDIFF=0.0001,
        # pymatgen has user set is_metal, where default is True and there
        # is no advanced logic. We just assume metal for now.
        ISMEAR=1,
        LDAU=False,
        KSPACING=0.35,  # !!! this is approximate to pymatgen
        multiple_keywords__smart_ldau="__remove__",
    )



# -*- coding: utf-8 -*-

import numpy

from simmate.toolkit import Structure


def clean_start_end_images(
    structure_start: Structure,
    structure_end: Structure,
    tolerance: float = 0.001,
):
    """
    Given two endpoint structures, this orders all sites to be the same and places
    the single moving site last in the structure.

    This function assumes all sites have an exact match in the second structure
    EXCEPT for a single site (which is the moving one).
    """

    # NOTES: code below is to help debugging
    # check if the order is already correct. If so, just skip this method
    # dists_orig = numpy.array(
    #     [s2.distance(s1) for s1, s2 in zip(structure_start, structure_end)]
    # )
    # if numpy.count_nonzero(dists_orig) == 1:
    #     return structure_start, structure_end

    base_structure = []

    starting_indicies = []
    ending_indicies = []
    for i1, site1 in enumerate(structure_start):
        found_match = False
        for i2, site2 in enumerate(structure_end):
            if site1.distance(site2) <= 0.001:
                base_structure.append(site1)
                found_match = True
                starting_indicies.append(i1)
                ending_indicies.append(i2)
                break
        # as a double-check, we keep track of the moving site
        if not found_match:
            moving_site_i = site1

    # find the indicies of the start/end moving site
    for i_start in range(len(structure_start)):
        if i_start not in starting_indicies:
            # make sure we have an agreement between the moving site found above and here
            assert structure_start[i_start] == moving_site_i
            break
    for i_end in range(len(structure_end)):
        if i_end not in ending_indicies:
            moving_site_e = structure_end[i_end]
            break

    structure_start_new = Structure.from_sites(base_structure + [moving_site_i])
    structure_end_new = Structure.from_sites(base_structure + [moving_site_e])

    # Make sure we successfully ordered the structure
    dists_new = numpy.array(
        [s2.distance(s1) for s1, s2 in zip(structure_start_new, structure_end_new)]
    )
    if not numpy.count_nonzero(dists_new) == 1:
        raise Exception("Failed to order supercell sites properly")

    return structure_start_new, structure_end_new



# -*- coding: utf-8 -*-

"""
This defines commands for managing your Simmate workflow engine. All commands are 
accessible through the "simmate engine" command.
"""

import typer

engine_app = typer.Typer(rich_markup_mode="markdown")


@engine_app.callback(no_args_is_help=True)
def engine():
    """
    A group of commands for starting up computational resources (Workers,
    Agents, and Clusters)
    """
    pass


@engine_app.command()
def start_schedules():
    """
    Starts the main process for periodic tasks in each app's "schedules" module.
    """
    from simmate.engine.scheduler import SimmateScheduler

    SimmateScheduler.start()


@engine_app.command()
def start_worker(
    nitems_max: int = None,
    timeout: float = None,
    close_on_empty_queue: bool = False,
    waittime_on_empty_queue: float = 1,
    tag: list[str] = ["simmate"],
):
    """
    Starts a Simmate Worker which will query the database for jobs to run

    By default this worker will run endlessly and not close.

    - `nitems_max`: the number of task run to submit before shutdown
    - `timeout`: the time (in seconds) after which this worker will stop running
    jobs and shutdown

    - `close_on_empty_queue`: whether the worker should shut down when the
    queue is empty

    - `waittime_on_empty_queue`: if the queue is empty, the time (in seconds)
    the worker should wait before checking the queue again

    - `tag`: tags to filter tasks by for submission. defaults to just 'simmate'.
    To provide multiple use `--tag example1 --tag example2` etc.

    """

    from simmate.engine import Worker

    worker = Worker(
        nitems_max,
        timeout,
        close_on_empty_queue,
        waittime_on_empty_queue,
        tag,  # this is actually "tags" --> a list of strings
    )
    worker.start()


@engine_app.command()
def start_singleflow_worker():
    """
    This starts a Simmate Worker that only runs one job and then shuts down. Also,
    if no job is available in the queue, it will shut down.

    Note: this can be acheived using the start-worker command too, but because
    this is such a common use-case, we include this command for convienence.
    It is the same as...

    `simmate engine start-worker --nitems-max 1 --close-on-empty-queue`
    """

    from simmate.engine import Worker

    Worker.run_singleflow_worker()


@engine_app.command()
def start_cluster(
    nworkers: int,
    type: str = "local",
    continuous: bool = False,
):
    """
    Starts many workers and (optionally) continuously submits new ones

    - `nworkers`: the number of workers to start

    - `cluster_type`: where to submit workers (either local or slurm)

    - `continuous`: whether to do a single submission of workers or hold nworkers
    at a steady-state number (runs endlessly)

    """

    from simmate.engine.execution.utilities import start_cluster

    start_cluster(
        nworkers=nworkers,
        cluster_type=type,
        continuous=continuous,
    )


@engine_app.command()
def error_summary():
    """
    Prints the shorthand error for all failed jobs
    """
    from simmate.engine.execution import SimmateExecutor

    SimmateExecutor.show_error_summary()


@engine_app.command()
def stats():
    """
    Prints a summary of different workitems and their status
    """
    from simmate.engine.execution import SimmateExecutor

    SimmateExecutor.show_stats()


@engine_app.command()
def stats_detail(tag: list[str] = [], recent: float = None):
    """
    Prints a summary of different workitems and their status
    - `tag`: tags to filter tasks by for submission. To provide multiple use
    `--tag example1 --tag example2` etc.

    - `recent`: show only items that have been updated in the last N hours.
    For example `--recent 24` limits to jobs editted in the last 24 hrs
    """
    from simmate.engine.execution import SimmateExecutor

    SimmateExecutor.show_stats_detail(
        tags=tag,
        recent=recent,
    )


@engine_app.command()
def workitems(tag: list[str] = [], status: str = None, recent: float = None):
    """
    Prints a table of all workitems and quick information about them

    - `tag`: tags to filter tasks by for submission. To provide multiple use
    `--tag example1 --tag example2` etc.

    - `status`: status of jobs to filter by. Options are P, R, F, E, and C

    - `recent`: show only items that have been updated in the last N hours.
    For example `--recent 24` limits to jobs editted in the last 24 hrs

    """
    from simmate.engine.execution import SimmateExecutor

    SimmateExecutor.show_workitems(
        tags=tag,
        status=status,
        recent=recent,
    )


@engine_app.command()
def delete_finished(confirm: bool = False):
    """
    Deletes all finished workitems from the database
    """
    from simmate.engine.execution import SimmateExecutor

    SimmateExecutor.delete_finished(confirm)


@engine_app.command()
def delete_all(confirm: bool = False):
    """
    Deletes all workitems from the database
    """
    from simmate.engine.execution import SimmateExecutor

    SimmateExecutor.delete_all(confirm)


@engine_app.command()
def delete(tag: list[str] = [], confirm: bool = False):
    """
    Deletes all workitems from the database that match the provided tags
    """
    from simmate.engine.execution import SimmateExecutor

    SimmateExecutor.delete(tags=tag, confirm=confirm)



# -*- coding: utf-8 -*-

from simmate.apps.warren_lab.workflows.relaxation.pbe import (
    Relaxation__Vasp__WarrenLabPbe,
)


class Relaxation__Vasp__WarrenLabPbesol(Relaxation__Vasp__WarrenLabPbe):
    """
    Runs a VASP relaxation calculation using Warren Lab HSE settings.
    """

    description_doc_short = "Warren Lab presets for HSE geometry optimization"

    # Tell vasp to use PBEsol instead of PBE
    _incar_updates = dict(GGA="PS")



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.matproj_scan import (
    Relaxation__Vasp__MatprojScan,
)


class StaticEnergy__Vasp__MatprojScan(Relaxation__Vasp__MatprojScan):
    """
    This task is a reimplementation of pymatgen's
    [MPScanStaticSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPScanStaticSet).
    """

    _incar_updates = dict(
        NSW=0,  # this is the main static energy setting
        LREAL=False,
        LORBIT=11,
        LVHAR=True,
        ISMEAR=-5,
    )



# -*- coding: utf-8 -*-

import logging
import subprocess

from simmate.engine.execution.cluster.base import Cluster


class SlurmCluster(Cluster):
    """
    Submits workers via a submit.sh file (in the working directory) and to
    a SLURM cluster.
    """

    @staticmethod
    def submit_job() -> int:
        """
        Submits a new job to the queue and returns the job id
        """
        process = subprocess.run(
            "sbatch submit.sh",
            shell=True,
            capture_output=True,
            text=True,
        )
        job_id = int(process.stdout.strip().split()[-1])
        return job_id

    @staticmethod
    def update_jobs_list(job_ids: list[int]) -> list[int]:
        """
        Given a list of job ids, it will check which ones are still running and
        which are finished. It will then return a list of the job id that are
        still running.
        """

        # OPTIMIZE: is there a way to queue a list of ids?
        still_running = []
        for job_id in job_ids:
            process = subprocess.run(
                f"squeue -j {job_id}",
                shell=True,
                capture_output=True,
                text=True,
            )
            # An error is return if the job is no longer in the queue.
            # If the job is still running, then two lines will be printed to
            # stdout AND the return code will be 0.
            if process.returncode == 0 and process.stdout.count("\n") == 2:
                still_running.append(job_id)
            else:
                logging.info(f"Slurm job {job_id} completed")

        return still_running



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.engine import Workflow


class ElectronicStructureWorkflow(Workflow):
    """
    Runs a static energy calculation followed by non-SCF calculations for
    band structure and density of states.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - static-energy.vasp.matproj
        - band-structure.vasp.matproj
        - density-of-states.vasp.matproj

    Note, these calculations are done using PBE, which is known to underestimate
    band gaps. For higher quality electronic structure calculations, you should
    use HSE instead.
    """

    use_database = False

    static_energy_workflow: Workflow = None
    band_structure_workflow: Workflow = None
    density_of_states_workflow: Workflow = None

    @classmethod
    def run_config(
        cls,
        structure,
        command: str = None,
        source: str = None,
        directory: Path = None,
        **kwargs,
    ):
        static_result = cls.static_energy_workflow.run(
            structure=structure,
            command=command,
            directory=directory / cls.static_energy_workflow.name_full,
            source=source,
            # For band-structures, unit cells should be in the standardized format
            standardize_structure="primitive",
        ).result()  # block until complete

        dos_state = cls.density_of_states_workflow.run(
            structure=static_result,
            command=command,
            directory=directory / cls.density_of_states_workflow.name_full,
        )

        bs_state = cls.band_structure_workflow.run(
            structure=static_result,
            command=command,
            directory=directory / cls.band_structure_workflow.name_full,
        )

        # block until complete
        dos_state.result()
        bs_state.result()

        # TODO:
        # workup dos_state.result() and bs_state.result() into single figure



# -*- coding: utf-8 -*-

from django import template

register = template.Library()


@register.filter
def sdfdoodle(obj):
    """
    Converts an sdf string to one that can be passed to the doodle_molecule fxn

    Example use:
    ``` html
    {% load sdfdoodle %}
    {{ my_object|sdfdoodle }}
    ```
    """
    return obj.replace("\n", "\\n")



# -*- coding: utf-8 -*-

import shutil
import subprocess

from simmate.toolkit import Composition, Structure
from simmate.toolkit.creators.structure.base import StructureCreator
from simmate.utilities import get_directory


class AirssStructure(StructureCreator):
    """
    Creates structures using the AIRSS software package.

    source: https://www.mtg.msm.cam.ac.uk/Codes/AIRSS
    tutorials: https://airss-docs.github.io/
    installation: https://airss-docs.github.io/getting-started/installation/

    Dev Notes:

    To install...
    It looks like we would only want their structure generation method, which
    is carried out via the buildcell script. They have a full page on its
    use: https://airss-docs.github.io/technical-reference/buildcell-manual/

    We first need to make a *.cell file with all of the inputs. The inputs are
    listed on their "Buildcell Manual" page, but the documentation is largely
    incomplete. After going through their examples, here's how I think we can
    make the *.cell file:
    ```
    #VARVOL=35
    #SPECIES=Si%NUM=1,O%NUM=2
    #MINSEP=1.0 Si-Si=3.00 Si-O=1.60 O-O=2.58
    ```

    Then to create a cif from these settings, we use the AIRSS structure
    converter (cabal):
    ```
    buildcell < test.cell | cabal cell cif > test.cif
    ```

    If a structure can't be made, this will continue indefinitely
    (MAXTIME doesnt fix this for some reason).
    """

    def __init__(
        self,
        composition: Composition,
        command: str = "buildcell",
    ):
        # check that the command exists
        if not shutil.which(command):
            raise Exception(
                "You must have AIRSS installed to use AirssStructure."
                "See https://airss-docs.github.io/getting-started/installation/"
            )
        self.command = command

        self.temp_dir = get_directory()  # leave empty to allow parallel

        # to setup the AIRSS creator, we need to make a *.cell file that
        # for example, a SiO2 file will look like... (NOTE - the # symbols
        # should be included in the file):
        # VARVOL=35
        # SPECIES=Si%NUM=1,O%NUM=2
        # MINSEP=1.0 Si-Si=3.00 Si-O=1.60 O-O=2.58
        self.cell_filename = self.temp_dir / "input.cell"
        with self.cell_filename.open("w") as file:
            # first write the VARVOL line using predicted volume
            volume = composition.volume_estimate()
            file.write("#VARVOL={}".format(volume) + "\n")

            # write the SPECIES line
            line = "#SPECIES="
            for element in composition:
                line += element.symbol + "%NUM=" + str(int(composition[element])) + ", "
            line = line[:-2]  # remove the final ', '
            # write the line
            file.write(line + "\n")

            # write the MINSEP line. For now I just assume 1 Angstrom
            file.write("#MINSEP=1.0" + "\n")

    def create_structure(self) -> Structure:
        # TODO: Maybe make an AIRSS app / S3 workflow
        output = subprocess.run(
            f"buildcell < {self.cell_filename} | cabal cell cif > AIRSS_output.cif",
            shell=True,
            capture_output=True,
            cwd=self.temp_dir,
        )

        # check to see if it was successful
        if output.returncode == 1:
            # AIRSS failed to make the structure
            return False

        # convert the cif file to pymatgen Structure
        filename = self.temp_dir / "AIRSS_output.cif"
        structure = Structure.from_file(filename)

        # and delete the cif file for cleanup
        filename.unlink()

        return structure



# -*- coding: utf-8 -*-

"""
This defines commands for managing your Simmate workflow engine. All commands are 
accessible through the "simmate engine" command.
"""

from pathlib import Path

import typer

utilities_app = typer.Typer(rich_markup_mode="markdown")


@utilities_app.callback(no_args_is_help=True)
def utilities():
    """
    A group of commands for various simple tasks (such as file handling)
    """
    pass


@utilities_app.command()
def archive_old_runs(
    directory: Path = Path.cwd(),
    time_cutoff: float = 3 * 7 * 24 * 60 * 60,  # equal to 3 weeks
):
    """
    Compresses old simmate-task-* folders to zip files

    - `directory`: the folder to search for old runs

    - `time_cutoff`: the time (in seconds) that a folder hasn't been editted in
    order to consider the run "old"
    """

    from simmate.utilities import archive_old_runs

    archive_old_runs(directory, time_cutoff)



# -*- coding: utf-8 -*-

from simmate.apps.warren_lab.inputs.potcar_mappings import HSE_POTCAR_MAPPINGS
from simmate.apps.warren_lab.workflows.relaxation.pbe import (
    Relaxation__Vasp__WarrenLabPbe,
)


class Relaxation__Vasp__WarrenLabHse(Relaxation__Vasp__WarrenLabPbe):
    """
    Runs a VASP relaxation calculation using Warren Lab HSE settings.
    """

    description_doc_short = "Warren Lab presets for HSE geometry optimization"
    # some potcars don't work with HSE in VASP 5.x.x. I change them here
    # in the hopes that they will work properly.
    potcar_mappings = HSE_POTCAR_MAPPINGS

    _incar_updats = dict(
        ALGO="Damped",  # We use Damped because it is the recommended setting by
        # by VASP (https://www.vasp.at/wiki/index.php/LHFCALC)
        HFSCREEN=0.2,
        ICHARG=1,
        LHFCALC=True,
        PRECFOCK="Fast",
        TIME=0.2,  # This is lower than the recommended setting (0.5) when using the
        # Damped tag because calculations struggle to converge.
        # VASP also suggests lowering it if convergence isn't reached so there
        # is precedence for this.
        VDW_S8=2.310,  # these three tags are necessary for IVDW 12 with HSE06
        VDW_A1=0.383,
        VDW_A2=5.685,
    )



# -*- coding: utf-8 -*-

"""
This defines commands for managing your Simmate configuration settings. All commands are 
accessible through the "simmate config" command.
"""

import logging
from pathlib import Path

import typer

config_app = typer.Typer(rich_markup_mode="markdown")


@config_app.callback(no_args_is_help=True)
def utilities():
    """
    A group of commands for managing Simmate settings
    """
    pass


@config_app.command()
def write(filename: Path = None):
    """
    Writes the final simmate settings to yaml file

    - `filename`: file name to write settings to
    """

    from simmate.configuration import settings

    settings.write_settings(filename=filename)


@config_app.command()
def show(user_only: bool = False):
    """
    Takes the final simmate settings and prints them in a yaml format that is
    easier to read.
    """
    from simmate.configuration import settings

    settings.show_settings(user_only=user_only)


@config_app.command()
def add(app_name: str):
    """
    Adds a specified Simmate app to the list of registered apps
    """

    from simmate.configuration import settings

    if app_name == "badelf":
        settings.add_apps_and_update(
            [
                "simmate.apps.configs.BaderConfig",
            ]
        )
    elif app_name == "bader":
        settings.add_apps_and_update(
            [
                "simmate.apps.configs.BaderConfig",
            ]
        )
    elif app_name == "clease":
        settings.add_apps_and_update(
            [
                "simmate.apps.configs.CleaseConfig",
            ]
        )
    elif app_name == "evolution":
        settings.add_apps_and_update(
            [
                "simmate.apps.configs.VaspConfig",  # TODO: deprec app dependency
                "simmate.apps.configs.EvolutionConfig",
            ]
        )
    elif app_name == "materials_project":
        settings.add_apps_and_update(
            [
                "simmate.apps.configs.VaspConfig",
                "simmate.apps.configs.BaderConfig",
                "simmate.apps.configs.MaterialsProjectConfig",
            ]
        )
    elif app_name == "quantum_espresso":
        settings.add_apps_and_update(
            [
                "simmate.apps.configs.QuantumEspressoConfig",
            ]
        )
    elif app_name == "vasp":
        settings.add_apps_and_update(
            [
                "simmate.apps.configs.VaspConfig",
            ]
        )
    elif app_name == "warren_lab":
        settings.add_apps_and_update(
            [
                "simmate.apps.configs.VaspConfig",
                "simmate.apps.configs.BaderConfig",
                "simmate.apps.configs.BadelfConfig",
                "simmate.apps.configs.WarrenLabConfig",
            ]
        )
    else:
        # user is giving a custom app
        settings.add_apps_and_update([app_name])


@config_app.command()
def update(config: str):
    """
    Updates Simmate settings using dot-notation
    """

    from simmate.configuration import settings

    config_cleaned = settings._parse_input(config)
    settings.write_updated_settings(config_cleaned)


@config_app.command()
def test(app_name: str):
    """
    Tests whether an app is properly configured
    """

    # OPTIMIZE: this code can be refactored and condensed. but it works for now

    if app_name == "badelf":
        from simmate.apps.bader.configuration import test_config

        passed = test_config()

    elif app_name == "bader":
        from simmate.apps.bader.configuration import test_config

        passed = test_config()

    elif app_name == "materials_project":
        from simmate.apps.materials_project.configuration import test_config

        passed = test_config()

    elif app_name == "quantum_espresso":
        from simmate.apps.quantum_espresso.configuration import test_config

        passed = test_config()

    elif app_name == "vasp":
        from simmate.apps.vasp.configuration import test_config

        passed = test_config()

    elif app_name == "warren_lab":
        from simmate.apps.warren_lab.configuration import test_config

        passed = test_config()

    else:
        logging.warning(f"Unknown app (or no tests exist): {app_name}")
        passed = False

    if not passed:
        raise typer.Exit(code=1)



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.mvl_neb_endpoint import (
    Relaxation__Vasp__MvlNebEndpoint,
)


class StaticEnergy__Vasp__MvlNebEndpoint(Relaxation__Vasp__MvlNebEndpoint):
    """
    Runs a VASP energy calculation using MIT Project settings, where some
    settings are adjusted to accomodate large supercells with defects.

    This is identical to relaxation/neb_endpoint, but just a single ionic step.

    You typically shouldn't use this workflow directly, but instead use the
    higher-level NEB workflows (e.g. diffusion/neb_all_paths or
    diffusion/neb_from_endpoints), which call this workflow for you.
    """

    _incar_updates = dict(
        IBRION=-1,  # (optional) locks everything between ionic steps
        NSW=0,
    )



# -*- coding: utf-8 -*-

import pytest

from simmate.command_line.engine import engine_app


@pytest.mark.django_db
def test_start_worker_cli(command_line_runner):
    # dump the database to json
    result = command_line_runner.invoke(
        engine_app,
        ["start-singleflow-worker"],
    )

    assert result.exit_code == 0

    # load the database to json
    result = command_line_runner.invoke(
        engine_app,
        ["start-worker", "--nitems-max", "1", "--close-on-empty-queue"],
    )
    assert result.exit_code == 0



# -*- coding: utf-8 -*-

import shutil
import subprocess

from simmate.toolkit import Composition, Structure
from simmate.toolkit.creators.structure.base import StructureCreator
from simmate.utilities import get_directory


class CalypsoStructure(StructureCreator):
    """
    Creates structures using the CALYPSO package

    This wrapper runs CALYPSO in 'split mode', where structures are simply
    created and then no further steps are taken.

    In the future, CALYPSO devs will hopefully have an accessible command
    # to just create structures. This will dramatically improve both the speed
    of this wrapper as well as the cleaniness of the code.

    source: http://www.calypso.cn/getting-calypso/


    ## Dev Notes

    For Ubuntu, I downloaded the x64 version of CALYPSO 7.0. Simply uncompress
    the download file and the executable is there. No need for any install.

    The following dependency is needed on ubuntu:
    ``` bash
    sudo apt-get install libomp-dev
    ```

    To generate structures without running any analysis/calcs, I simply need
    to run calypso in split mode:
    ```
    Split = T
    ```

    All the structures are put into POSCAR files named POSCAR_1, POSCAR_2, ... POSCAR_N
    """

    def __init__(
        self,
        composition: Composition,
        # location of the calypso.x file
        command: str = "calypso.x",
    ):
        if not shutil.which(command):
            raise Exception("You must have CALYPSO installed and in the PATH.")

        self.command = command

        # Because of bug with calypso's poscar files, I need to save this
        self.comp = " ".join([element.symbol for element in composition])

        # set all of the USPEX options and format them the way they'd like
        # There are many  more options (such as for 2D materials) that can be added
        calypso_options = {
            # number of unique elements
            "NumberOfSpecies": len(composition),
            # This is a list of atomic symvols (for example, MgSiO3 is 'Mg Si O')
            "NameOfAtoms": " ".join([element.symbol for element in composition]),
            # This is a list of reduced species count
            "NumberOfAtoms": " ".join(
                [
                    str(int(composition.reduced_composition[element]))
                    for element in composition
                ]
            ),
            # factor of formula unit -- for example Mg3Si3O9 would have 3 (vs MgSiO3)
            "NumberOfFormula": int(
                composition.num_atoms / composition.reduced_composition.num_atoms
            ),
        }

        # CALYPSO's auto volume prediction doesn't look to work very well.
        # It's either that, or the distance checks are too strict.
        # Therefore, let's predict volume on our own and input it.
        volume = composition.volume_estimate(packing_factor=2)
        calypso_options.update({"Volume": volume})

        # CALYPSO says they have a default value for distances, but it doesn't
        # look like they work. So I am going to end up making a distance
        # matrix here.
        dm = composition.distance_matrix_estimate()
        dm_str = ""
        for row in dm:
            for val in row:
                dm_str += str(val) + " "
            dm_str += "\n"
        dm_str = dm_str[:-2]  # remove the final \n
        calypso_options.update({"DistanceOfIon": dm_str})

        # Now format this input string with our options. The only remaining
        # input to set is the number of structures to generate (NUM_STRUCTURES)
        self.calypso_input = INPUT_TEMPLATE.format(**calypso_options)

    def create_structures(self, n: int) -> list[Structure]:
        # See my comments above on why this atypical function exists...
        # (it's much faster than calling USPEX each new struct)

        temp_dir = get_directory()  # leave empty to allow parallel

        # make the INPUT.txt file with n as our NUM_STRUCTURES to make
        # write it and close immediately
        input_file = temp_dir / "input.dat"
        with input_file.open("w") as file:
            file.writelines(self.calypso_input.replace("NUM_STRUCTURES", str(n)))

        # now let's have CALYPSO run and make the structures
        subprocess.run(
            self.command,
            shell=True,
            capture_output=True,
            cwd=temp_dir,
        )

        # All the structures are as POSCAR files. The files are
        #   POSCAR_1, POSCAR_2, ... POSCAR_n...
        # Let's iterate through these and pool them into a list
        structures = []
        for poscar_file in temp_dir.iterdir():
            if not poscar_file.stem.startswith("POSCAR_"):
                continue

            # ----------------------------
            # BUG WITH CALYPSO...
            # They don't add the atom types to the POSCAR... wtf
            # I need to do that manually here
            with poscar_file.open("r") as file:
                lines = file.readlines()
            # add the composition line
            lines.insert(5, self.comp + "\n")
            # write the updated file
            with poscar_file.open("w") as file:
                file.writelines(lines)
            # ----------------------------

            # now we can load the POSCAR and add it to our list
            structure = Structure.from_file(poscar_file)
            structures.append(structure)

        # delete the directory
        shutil.rmtree(temp_dir)

        # return the list of pymatgen Structure objects that we've made
        return structures

    def create_structure(self) -> Structure:
        # call the create_structures() function and tell it to create
        # just one structure
        structure = self.create_structures(1)[0]

        return structure


# Here we setup the input file to create CALYPSO structures.
# Note NumberOfFormula is input twice becuase we want a fixed min/max
INPUT_TEMPLATE = """
NumberOfSpecies = {NumberOfSpecies}
NameOfAtoms = {NameOfAtoms}
NumberOfAtoms = {NumberOfAtoms}
NumberOfFormula = {NumberOfFormula} {NumberOfFormula}
PopSize = NUM_STRUCTURES
Command = echo SKIP THIS
Split = T
Volume = {Volume}
@DistanceOfIon
{DistanceOfIon}
@End
"""



# -*- coding: utf-8 -*-

import shutil
from pathlib import Path

from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from simmate.apps.vasp.inputs import Incar, Kpoints, Poscar, Potcar
from simmate.configuration import settings
from simmate.engine import S3Workflow
from simmate.toolkit import Structure


def get_default_parallel_settings():
    """
    We load the user's default parallel settings for VASP, which optionally
    configure NCORE and KPAR for all workflows.

    These settings should not effect the calculation result in any way
    (only how fast it completes), but they are still selected based on
    the computer specs and what runs fastest on it.
    """
    parallel_settings = settings.vasp.parallelization

    if parallel_settings.ncore is not None or parallel_settings.kpar is not None:
        raise NotImplementedError(
            "NCORE and KPAR configuration settings are still under development"
        )
    else:
        return {}


class VaspWorkflow(S3Workflow):
    _parameter_methods = S3Workflow._parameter_methods + ["_get_clean_structure"]

    required_files = ["INCAR", "POTCAR", "POSCAR"]
    exlcude_from_archives = ["POTCAR"]

    command: str = "vasp_std > vasp.out"
    """
    The command to call vasp, which is typically vasp_std. To ensure error
    handlers work properly, make sure your command has "> vasp.out" at the end.
    """
    # TODO: add support for grabbing a user-set default from their configuration
    # TODO: add auto check for vasp.out ending

    # -------------------------------------------------------------------------

    # INCAR configuration

    @classmethod
    @property
    def incar(cls) -> dict:
        """
        The INCAR configuration settings that will be used for this workflow.

        Note: do not confuse with `_incar` and `_incar_updates`, which are
        used to determine the final settings given by this `incar` property.
        """
        # OPTIMIZE: I'd like this to be cached
        if not cls._incar_updates:
            return cls._incar
        else:
            # grab the parent workflow to use as the base incar. The workflow
            # may have multiple parent classes (mix-ins), so we grab __bases__
            # instead of __base__
            parent_flows = cls.__bases__
            if len(parent_flows) == 1:
                parent_incar = cls.__base__.incar
            else:
                parent_incars_found = 0
                for parent_flow in parent_flows:
                    if parent_flow.incar:
                        parent_incar = parent_flow.incar
                        parent_incars_found = +1
                        # we do not break the loop in case there are errors, which
                        # we check for below
                if parent_incars_found > 1:
                    raise Exception(
                        "Your VaspWorkflow is inheriting from more than workflow "
                        "with valid `incar` settings. Therefore, it is not clear "
                        "which settings should be inherited in your subclass."
                    )
                elif parent_incars_found == 0:
                    raise Exception(
                        "Only use `_incar_updates` when the parent workflow has a "
                        "valid `incar` config set."
                    )
            # Note we always use copy() in the methods below because we
            # are modifying all values in-place
            updates = cls._incar_updates.copy()
            final_incar = parent_incar.copy()

            for key, value in updates.items():
                # Check if there is a modifier version of the key that
                # needs removed.
                # For example, a new "EDIFF" would replace "EDIFF__per_atom"
                for original_key in parent_incar.keys():
                    if original_key.split("__")[0] == key.split("__")[0]:
                        final_incar.pop(original_key)
                # OPIMIZE: maybe move this logic to the Incar class

                # now set the updated value
                if value == "__remove__":
                    final_incar.pop(value, None)
                else:
                    final_incar[key] = value

            return final_incar

    _incar: dict = {}
    """
    This sets the base vasp settings from a dictionary. This is the one thing
    you *must* set when subclassing VaspWorkflow. An example is:
    
    ``` python
      _incar = dict(NSW=0, PREC="Accurate", KSPACING=0.5)
    ```
    
    Note: you should refer to `incar` instead of this attribute for the final
    workflow settings of this class.
    """

    _incar_updates: dict = {}
    """
    When subclassing a subclass of `VaspWorkflow` (i.e. inheriting INCAR settings),
    this attribute helps to set which INCARs are changed relative to the parent
    workflow.
    """

    # -------------------------------------------------------------------------

    # TODO: add options for poscar formation
    # add_selective_dynamics=False
    # add_velocities=False
    # significant_figures=6 --> rounding issues? what's the best way to do this?

    kpoints = None
    """
    (experimental feature)
    The KptGrid or KptPath generator used to create the KPOINTS file. Note,
    this attribute is optional becuase VASP supports setting Kpts by adding
    KSPACING to the INCAR. If KSPACING is set in the INCAR, we ignore whatever 
    is set here.
    """
    # TODO - KptGrid is just a float for now, so there's no typing here.

    functional: str = None
    """
    This directs which Potcar files to grab. You would set this to a string
    of what you want, such as "PBE", "PBE_GW", or "LDA".
    """

    potcar_mappings: dict = None
    """
    This is an optional parameter to override Simmate's default selection of
    potentials based off of the functional chosen. The defaults are located
    in `simmate.apps.vasp.inputs.potcar_mappings`. You can supply your
    own mapping dictionary or update the specific potentials you'd like. 
    For example:
    
    ``` python
      from simmate.apps.vasp.inputs import PBE_POTCAR_MAPPINGS
      potcar_mappings = PBE_POTCAR_MAPPINGS.copy()  # don't forget to copy!
      potcar_mappings.update({"C": "C_h"})  # if you wish to update any
    ```
    
    or if you only use Carbon and don't care about other elements...
    
    ``` python
      potcar_mappings = {"C": "C_h"}
    ```
    
    Read more on this inside the Potcar class and be careful with updating!
    """

    standardize_structure: str | bool = False
    """
    In some cases, we may want to standardize the structure during our setup().
    
    This means running symmetry analysis on the structure in order to reduce
    the symmetry and also converting it to some standardized form. There
    are three different forms to choose from and thus 3 different values that
    `standardize_structure` can be set to:
    
    - `primitive`: for the standard primitive unitcell
    - `conventional`: for the standard conventional unitcell
    - `primitive-LLL`: for the standard primitive unitcell that is then LLL-reduced
    - `False`: this is the default and will disable this feature
    
    We recommend using `primitive-LLL` when the smallest possible and most cubic
    unitcell is desired.
    
    We recommend using `primitive` when calculating band structures and 
    ensuring we have a standardized high-symmetry path. Note, existing band
    structure workflows automatically adjust this value for you.
    
    To control the tolerances used to symmetrize the structure, you can use the
    symmetry_precision and angle_tolerance attributes.
    
    By default, no standardization is applied.
    """

    symmetry_precision: float = 0.01
    """
    If standardize_structure=True, then this is the cutoff value used to determine
    if the sites are symmetrically equivalent. (in Angstroms)
    """

    angle_tolerance: float = 0.5
    """
    If standardize_structure=True, then this is the cutoff value used to determine
    if the angles between sites are symmetrically equivalent. (in Degrees)
    """

    @classmethod
    def _get_clean_structure(
        cls,
        structure: Structure,
        standardize_structure: str | bool = None,
        symmetry_precision: float = None,
        angle_tolerance: float = None,
        **kwargs,
    ) -> Structure:
        """
        Uses the class attribute for `standardize_structure`. If this is set
        to any of the listed modes, then the structure unitcell is converted
        using the proper methods.

        Note, this method is typically called within `setup` before any input
        files are written. You should never have to call it directly.
        """

        # See if these values were provided, or default to class attribute.
        # We rename this to "standardize mode" to accurately descibe the
        # variable if it is set
        standardize_mode = standardize_structure or cls.standardize_structure
        symmetry_precision = symmetry_precision or cls.symmetry_precision
        angle_tolerance = angle_tolerance or cls.angle_tolerance

        # if standardize_structure is not requested, then we just return the
        # orignal input structure.
        if not standardize_mode:
            return structure

        # Make sure the user
        if standardize_mode not in ["primitive-LLL", "conventional", "primitive"]:
            raise Exception(
                f"Standardization mode {standardize_mode} is not supported."
            )

        # Both cleaning steps start with looking at the symmetry analysis
        sym_finder = SpacegroupAnalyzer(
            structure,
            symprec=symmetry_precision,
            angle_tolerance=angle_tolerance,
        )

        # If requested, we convert to the LLL-reduced unit cell, which aims to
        # be as cubic as possible.
        if standardize_mode == "primitive-LLL":
            structure_prim = sym_finder.get_primitive_standard_structure()
            structure_cleaned = structure_prim.copy(sanitize=True)
            return structure_cleaned

        # For band structures, we need to make sure the structure is in the
        # standardized primitive form.
        if standardize_mode == "primitive":
            structure_cleaned = sym_finder.get_primitive_standard_structure(
                international_monoclinic=False,
            )
            # check for pymatgen bugs here
            check_for_standardization_bugs(structure, structure_cleaned)
            return structure_cleaned

        # lastly is the conventional standard, which is used the least.
        if standardize_mode == "conventional":
            structure_cleaned = sym_finder.get_conventional_standard_structure()
            return structure_cleaned

    @classmethod
    def setup(cls, directory: Path, structure: Structure, **kwargs):
        # run cleaning and standardizing on structure (based on class attributes)
        structure_cleaned = cls._get_clean_structure(structure, **kwargs)

        # write the poscar file
        Poscar.to_file(structure_cleaned, directory / "POSCAR")

        # Combine our base incar settings with those of our parallelization settings
        # and then write the incar file
        incar = Incar(**cls.incar)
        incar.to_file(
            filename=directory / "INCAR",
            structure=structure_cleaned,
        )

        # if KSPACING is not provided in the incar AND kpoints is attached to this
        # class instance, then we write the KPOINTS file
        if cls.kpoints and ("KSPACING" not in cls.incar):
            Kpoints.to_file(
                structure_cleaned,
                cls.kpoints,
                directory / "KPOINTS",
            )

        # write the POTCAR file
        Potcar.to_file_from_type(
            structure_cleaned.composition.elements,
            cls.functional,
            directory / "POTCAR",
            cls.potcar_mappings,
        )

    @classmethod
    def setup_restart(cls, directory: Path, **kwargs):
        """
        From a working directory of a past calculation, sets up for the calculation
        to be restarted. For relaxations/dynamics this involved just copying
        the poscar to the contcar.
        """

        # establish filenames
        poscar_filename = directory / "POSCAR"
        poscar_orig_filename = directory / "POSCAR_original"
        contcar_filename = directory / "CONTCAR"
        stopcar_filename = directory / "STOPCAR"

        # TODO:
        # make an archive of the directory before we start editting files
        # make_error_archive(directory)
        # add these changes to the simmate_corrections.csv

        # delete the stopcar if it exists
        if stopcar_filename.exists():
            stopcar_filename.unlink()

        # copy poscar to a new file
        shutil.move(poscar_filename, poscar_orig_filename)

        # then CONTCAR over to the POSCAR
        shutil.move(contcar_filename, poscar_filename)

    @classmethod
    def get_config(cls):
        """
        Grabs the overall settings from the class. This is useful for printing out
        settings for users to inspect.
        """
        return {
            key: getattr(cls, key)
            for key in [
                "__module__",
                "functional",
                "incar",
                "potcar_mappings",
            ]
        }


def check_for_standardization_bugs(
    structure_original: Structure,
    structure_new: Structure,
):
    # In pymatgen, they include this code with the standardization of their
    # structures because there were several bugs in the past and they want to
    # double-check themselves. I'm still using their code to standardize
    # my structures, so I should make this check too.

    vpa_old = structure_original.volume / structure_original.num_sites
    vpa_new = structure_new.volume / structure_new.num_sites

    if abs(vpa_old - vpa_new) / vpa_old > 0.02:
        raise ValueError(
            "Standardizing failed! Volume-per-atom changed... "
            f"old: {vpa_old}, new: {vpa_new}"
        )

    sm = StructureMatcher()
    if not sm.fit(structure_original, structure_new):
        raise ValueError("Standardizing failed! Old structure doesn't match new.")



# -*- coding: utf-8 -*-

from pathlib import Path

from pymatgen.analysis.diffusion.neb.pathfinder import DistinctPathFinder as PymatgenDPF


class DistinctPathFinder(PymatgenDPF):
    def write_all_migration_hops(self, directory: Path):
        # We write all the path files so users can visualized them if needed
        filename = directory / "all_migration_hops.cif"
        self.write_all_paths(str(filename), nimages=10)
        migration_hops = self.get_paths()
        for i, migration_hop in enumerate(migration_hops):
            number = str(i).zfill(2)  # converts numbers like 2 to "02"
            # the files names here will be like "migration_hop_02.cif"
            migration_hop.write_path(
                str(directory / f"migration_hop_{number}.cif"),
                nimages=10,  # this is just for visualization
            )



# -*- coding: utf-8 -*-

"""

This file is for pulling MPDS data into the Simmate database. 

MPDS provides a python package to make downloading their data very easy. This 
officially supported and maintained at https://github.com/mpds-io/mpds_client.
For now the package is only available via a pip install. And to help with using
this package, you should read through their guide and examples here:    
https://mpds.io/developer/#Client-library

NOTE: I'm currently running into permission issues with MPDS, and will return to
this provider after talking with their team.

"""

from django.db import transaction
from mpds_client import MPDSDataRetrieval, MPDSDataTypes
from rich.progress import track

from simmate.database import connect
from simmate.database.third_parties.mpds import MpdsStructure
from simmate.toolkit import Structure


@transaction.atomic
def load_all_structures(api_key="9WPgw8xYqYcB91ylff1CogIaXd4MGq6RoEZjwC1L0fh66uMw"):
    # !!! Remove key in production

    # connect to the database
    client = MPDSDataRetrieval(api_key)

    # By default, only structures from peer-reviewed papers are shown, but we
    # can switch this to include MPDS's in-house machine learning and first
    # principles structures too.
    # !!! only the ab initio and machine learning datasets are free and open
    # so I have to limit myself to those for now.
    #   client.dtype = MPDSDataTypes.ALL
    client.dtype = MPDSDataTypes.AB_INITIO

    # get_data returns a list, while get_dataframe will give us a pandas dataframe
    # I just use a list for now.
    # !!! This doesn't behave the way they claim... it definitely doesn't return
    # JSON. I'm also only ever getting hits for P-type ids (properties), when I
    # want S-type (structures)
    data = client.get_data(
        search={
            "elements": "Al-N",  # !!! for testing
            # "props": "atomic structure",
        },
        # fields={
        #     "S": [
        #         # "phase_id",
        #         # "entry",
        #         'chemical_formula',
        #         # "cell_abc",
        #         # "sg_n",
        #         # "basis_noneq",
        #         # "els_noneq",
        #     ]
        # },
    )



# -*- coding: utf-8 -*-

import os
import shutil

# This will be added back once I go through and handle warnings within context
# import warnings
from pathlib import Path

from simmate.apps.badelf.core.badelf import BadElfToolkit
from simmate.engine import Workflow
from simmate.toolkit import Structure

# This file contains workflows for performing Bader and BadELF. Parts of the code
# use the Henkelman groups algorithm for Bader analysis:
# (http://theory.cm.utexas.edu/henkelman/code/bader/).


class BadElfBase(Workflow):
    """
    Controls a Badelf analysis on a pre-ran VASP calculation.
    This is the base workflow that all analyses that run BadELF
    are built from. Note that for more in depth analysis, it may be more
    useful to use the BadElfToolkit class.
    """

    use_database = False

    @classmethod
    def run_config(
        cls,
        source: dict = None,
        directory: Path = None,
        cores: int = None,
        find_electrides: bool = True,
        electride_finder_cutoff: float = 0.5,  # This is somewhat arbitrarily set
        algorithm: str = "badelf",
        electride_connection_cutoff: float = 0,
        check_for_covalency: bool = True,
        **kwargs,
    ):
        # make a new directory to run badelf algorithm in and copy necessary files.
        badelf_directory = directory / "badelf"
        try:
            os.mkdir(badelf_directory)
        except:
            pass
        files_to_copy = ["CHGCAR", "ELFCAR", "POTCAR"]
        for file in files_to_copy:
            shutil.copy(directory / file, badelf_directory)

        # Get the badelf toolkit object for running badelf.
        badelf_tools = BadElfToolkit.from_files(
            directory=badelf_directory,
            find_electrides=find_electrides,
            algorithm=algorithm,
        )
        # Set options and run badelf.
        if not check_for_covalency:
            badelf_tools.check_for_covalency = False
        badelf_tools.electride_finder_cutoff = electride_finder_cutoff
        badelf_tools.electride_connection_cutoff = electride_connection_cutoff
        results = badelf_tools.results
        badelf_tools.write_results_csv()
        return results


class VaspBadElfBase(Workflow):
    """
    Runs a VASP DFT calculation followed by a BadELF analysis. This is the base
    workflow that all analyses that run DFT and BadELF are built from.
    """

    use_database = False
    static_energy_workflow = None  # This should be defined in the inheriting class
    badelf_workflow = None  # This should be defined in the inheriting class
    # usually as the BadElfAnalyisis__WarrenLab__badelf workflow

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        command: str = None,
        source: dict = None,
        directory: Path = None,  # we default to the current directory
        cores: int = None,
        find_electrides: bool = True,
        electride_finder_cutoff: float = 0.5,  # This is somewhat arbitrarily set
        algorithm: str = "badelf",
        electride_connection_cutoff: float = 0,
        check_for_covalency: bool = True,
        **kwargs,
    ):
        # Run the dft calculation. This workflow should be set as something
        # that already saves to a database table.
        static_energy_directory = directory / "static_energy"
        static_energy_result = cls.static_energy_workflow.run(
            structure=structure,
            command=command,
            source=source,
            directory=static_energy_directory,
        ).result()

        # run the badelf analysis in the same directory as the DFT. This workflow
        # should be set to something that already saves to a database table.

        badelf_result = cls.badelf_workflow.run(
            source=source,
            directory=static_energy_directory,
            cores=cores,
            find_electrides=find_electrides,
            electride_finder_cutoff=electride_finder_cutoff,
            algorithm=algorithm,
            electride_connection_cutoff=electride_connection_cutoff,
            check_for_covalency=check_for_covalency,
            # copy_previous_directory=True,
        ).result()



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.apps.quantum_espresso.inputs import PwscfInput
from simmate.apps.quantum_espresso.inputs.k_points import Kpoints
from simmate.apps.quantum_espresso.inputs.potentials_sssp import (
    SSSP_PBE_EFFICIENCY_MAPPINGS,
    SSSP_PBE_PRECISION_MAPPINGS,
)
from simmate.configuration import settings
from simmate.engine import S3Workflow
from simmate.toolkit import Structure
from simmate.utilities import get_docker_command


# TODO: add StructureInputWorkflow mixin which can be made from VaspWorkflow class
class PwscfWorkflow(S3Workflow):
    required_files = ["pwscf.in"]

    command: str = settings.quantum_espresso.default_command
    """
    The command to call PW-SCF, which is typically `pw.x`.
    
    The typical default is "pw.x < pwscf.in > pw-scf.out"
    """

    # -------------------------------------------------------------------------

    # We set each section of PW-SCF's input parameters as a class attribute
    # https://www.quantum-espresso.org/Doc/INPUT_PW.html

    control: dict = {}
    """
    key-value pairs for the `&CONTROL` section of `pwscf.in`
    """

    system: dict = {}
    """
    key-value pairs for the `&SYSTEM` section of `pwscf.in`
    """

    electrons: dict = {}
    """
    key-value pairs for the `&ELECTRONS` section of `pwscf.in`
    """

    ions: dict = {}
    """
    key-value pairs for the `&IONS` section of `pwscf.in`
    """

    cell: dict = {}
    """
    key-value pairs for the `&CELL` section of `pwscf.in`
    """

    fcp: dict = {}
    """
    key-value pairs for the `&FCP` section of `pwscf.in`
    """

    rism: dict = {}
    """
    key-value pairs for the `&RISM` section of `pwscf.in`
    """

    @classmethod
    @property
    def full_settings(cls) -> dict:
        # TODO: consider making this use PwscfInput class
        return dict(
            control=cls.control,
            system=cls.system,
            electrons=cls.electrons,
            ions=cls.ions,
            fcp=cls.fcp,
            rism=cls.rism,
        )

    # -------------------------------------------------------------------------

    psuedo_mappings_set: str = None
    """
    Indicates which psuedopotentials mappings to use (in the `psuedo_mappings` attribute).
    Can be either 'SSSP_PBE_PRECISION' or 'SSSP_PBE_EFFICIENCY'
    """

    @classmethod
    @property
    def psuedo_mappings(cls) -> dict:
        if cls.psuedo_mappings_set == "SSSP_PBE_PRECISION":
            return SSSP_PBE_PRECISION_MAPPINGS
        elif cls.psuedo_mappings_set == "SSSP_PBE_EFFICIENCY":
            return SSSP_PBE_EFFICIENCY_MAPPINGS
        else:
            raise Exception(
                f"Unknown psuedo_mappings_set provided: {cls.psuedo_mappings_set}"
            )

    # -------------------------------------------------------------------------

    k_points: dict = {}
    """
    Configuration for generating the k-points grid for each input. This
    config is passed to `Kpoints.from_dynamic`
    """

    # -------------------------------------------------------------------------

    @classmethod
    def setup(cls, directory: Path, structure: Structure, **kwargs):
        # run cleaning and standardizing on structure (based on class attributes)
        # TODO: structure_cleaned = cls._get_clean_structure(structure, **kwargs)

        input_config = PwscfInput(
            structure=structure,
            kpoints=Kpoints.from_dynamic(
                k_points=cls.k_points,
                structure=structure,
            ),
            psuedo_mappings=cls.psuedo_mappings,
            control=cls.control,
            system=cls.system,
            electrons=cls.electrons,
            ions=cls.ions,
            cell=cls.cell,
            fcp=cls.fcp,
            rism=cls.rism,
        )
        input_config.to_file(directory / "pwscf.in")

    @classmethod
    def get_final_command(
        cls, command: str = None, directory: Path = None, **kwargs
    ) -> str:
        # EXPERIMENTAL - some of this functionality will likely move to S3Workflow
        if settings.quantum_espresso.docker.use == True:
            final_command = get_docker_command(
                image=settings.quantum_espresso.docker.image,
                entrypoint=command,
                volumes=[
                    f"{str(directory)}:/qe_calc",
                    f"{str(settings.quantum_espresso.psuedo_dir)}:/potentials",
                ],
            )
        breakpoint()
        return final_command

    # -------------------------------------------------------------------------



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.apps.quantum_espresso.inputs.potentials_sssp import (
    SSSP_PBE_EFFICIENCY_MAPPINGS,
    SSSP_PBE_PRECISION_MAPPINGS,
)
from simmate.configuration import settings
from simmate.toolkit import Structure


def keyword_modifier_pseudo_dir(structure: Structure, confirm_auto: bool) -> Path:
    """
    If `pseudo_dir__auto=True` is set, the user is requesting the default directory
    for psuedopotentials
    """
    assert confirm_auto
    if not settings.quantum_espresso.docker.enable:  # used in the majority of cases
        return settings.quantum_espresso.psuedo_dir
    else:
        # the docker image will have a volume mapped to here
        return "/potentials"


def keyword_modifier_nat(structure: Structure, confirm_auto: bool) -> int:
    """
    If `nat__auto=True` is set, the user wants nat (number of atoms) set automatically
    """
    assert confirm_auto
    return len(structure)


def keyword_modifier_ntyp(structure: Structure, confirm_auto: bool) -> int:
    """
    If `ntyp__auto=True` is set, the user wants ntyp (number of species) set automatically
    """
    assert confirm_auto
    return len(structure.composition)


def keyword_modifier_ecutwfc(structure: Structure, mode: str) -> float:
    """
    If `ecutwfc__auto=True` is set, the user wants ecutwfc (Kinetic energy cutoff (Ry)
    for wavefunctions) set automatically. To do this, we look at all elements
    and their mapped psuedos. The mappings indicate a suggested value and we use
    the maximum across all elements.
    """
    if mode == "efficiency":
        mappings = SSSP_PBE_EFFICIENCY_MAPPINGS
    elif mode == "precision":
        mappings = SSSP_PBE_PRECISION_MAPPINGS
    else:
        raise Exception(f"Unknown mode set for ecutwfc__auto: {mode}")

    return max(
        [mappings[element.symbol]["cutoff_wfc"] for element in structure.composition]
    )


def keyword_modifier_ecutrho(structure: Structure, mode: str) -> float:
    """
    If `ecutrho__auto=True` is set, the user wants ecutrho (Kinetic energy cutoff (Ry)
    for charge density and potential) set automatically. To do this, we look at
    all elements and their mapped psuedos. The mappings indicate a suggested
    value and we use the maximum across all elements.
    """
    if mode == "efficiency":
        mappings = SSSP_PBE_EFFICIENCY_MAPPINGS
    elif mode == "precision":
        mappings = SSSP_PBE_PRECISION_MAPPINGS
    else:
        raise Exception(f"Unknown mode set for ecutrho__auto: {mode}")

    return max(
        [mappings[element.symbol]["cutoff_rho"] for element in structure.composition]
    )



from simmate.apps.warren_lab.workflows import Relaxation__Vasp__Staged


class Relaxation__Vasp__StagedCluster(Relaxation__Vasp__Staged):
    subworkflow_names = [
        "relaxation.vasp.cluster-low-q",
        "relaxation.vasp.cluster-high-q",
        "static-energy.vasp.cluster-high-qe",
    ]



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.matproj import (
    Relaxation__Vasp__Matproj,
)


class Elastic__Vasp__Mvl(Relaxation__Vasp__Matproj):
    """
    This task is a reimplementation of pymatgen's
    [MVLElasticSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MVLElasticSet).
    """

    _incar_updates = dict(
        IBRION=6,
        NFREE=2,
        POTIM=0.015,  # pymatgen uses this as a parameter
        NPAR="__remove__",  # pymatgen forcibly removes NPAR
    )



# -*- coding: utf-8 -*-

import logging
import time
import traceback

import cloudpickle  # needed to serialize Prefect workflow runs and tasks
from django.db import transaction
from rich import print

from simmate.engine.execution.database import WorkItem

# This string is just something fancy to display in the console when a worker
# starts up.
# This uses "Small Slant" from https://patorjk.com/software/taag/
HEADER_ART = r"""
=====================================================================
   _____                  __        _      __         __
  / __(_)_ _  __ _  ___ _/ /____   | | /| / /__  ____/ /_____ ____
 _\ \/ /  ' \/  ' \/ _ `/ __/ -_)  | |/ |/ / _ \/ __/  '_/ -_) __/
/___/_/_/_/_/_/_/_/\_,_/\__/\__/   |__/|__/\___/_/ /_/\_\\__/_/

=====================================================================
"""


class SimmateWorker:
    """
    The default worker that connect to Simmate database for workflows submitted
    via the `run_cloud` method.
    """

    # Consider making this a database object so that we can track workers in
    # the UI and know how many are running. If linked to the database, we could
    # even send an update for the worker to shut down.

    # Ideally, this worker would involve multiple threads threads going. One
    # thread would update the queue database with a "heartbeat" to let it know
    # that it is still working on tasks. The other thread will run the given
    # workitems in serial. I could even allow for more threads so that this
    # worker can run multiple workitems at once and in parallel.
    # However, if this level of implementation is needed, we should instead
    # switch to using Prefect, which has it built in.

    def __init__(
        self,
        # limit of tasks and lifetime of the worker
        nitems_max: int = None,
        timeout: float = None,
        # wait_on_timeout=False, # TODO
        # settings on what to do when the queue is empty
        close_on_empty_queue: bool = False,
        waittime_on_empty_queue: float = 15,
        tags: list[str] = ["simmate"],  # should default be empty...?
    ):
        """
        Configures a worker that connects to the default executor backend.
        By default the worker will run endlessly.

        #### Parameters:

        - `nitems_max`:
            The maximum number of workitems to run before closing down
            if no limit was set, we can go to infinity.

        - `timeout`:
            Don't start a new workitem after this time. The worker will be shut down.
            if no timeout was set, use infinity so we wait forever.

        - `close_on_empty_queue`:
            whether to close if the queue is empty

        - `waittime_on_empty_queue`:
            if the queue is found to be empty, check the queue again after
            this time sleeping.

        - `tags`:
            the tags to query tasks for. If no tags were given, the worker will
            query for tasks that have NO tags

        """
        self.tags = tags
        self.nitems_max = nitems_max or float("inf")
        self.timeout = timeout or float("inf")
        self.close_on_empty_queue = close_on_empty_queue
        self.waittime_on_empty_queue = waittime_on_empty_queue

        # whether to wait on the running workitems to finish before shutting down
        # the timedout worker.
        # self.wait_on_timeout = wait_on_timeout # # TODO

    def start(self):
        """
        Starts the worker process to begin working through WorkItems
        """

        # print the header in the console to let the user know the worker started
        print("[bold cyan]" + HEADER_ART)

        # loggin helpful info
        logging.info(f"Starting worker with tags {list(self.tags)}")

        # establish starting point for the worker
        time_start = time.time()
        ntasks_finished = 0

        # Loop endlessly until one of the following happens...
        #   the timeout limit is hit
        #   the queue is empty
        #   the nitems limit is hit
        while True:
            # check for timeout before starting a new workitem and exit
            # if we've hit the limit.
            if (time.time() - time_start) > self.timeout:
                # TODO - check wait_on_timeout if running in parallel.
                logging.info(
                    "The time-limit for this worker has been hit. Shutting down."
                )
                return

            # check the number of jobs completed so far, and exit if we hit
            # the limit
            if ntasks_finished >= self.nitems_max:
                logging.info(
                    f"Maximum number of WorkItems reached ({self.nitems_max}). "
                    "Shutting down."
                )
                return

            # check the length of the queue and while it is empty, we want to
            # loop. The exception of looping endlessly is if we want the worker
            # to shutdown instead.
            while self.queue_size() == 0:
                # if it is empty, we want to sleep for a little and check again
                time.sleep(self.waittime_on_empty_queue)

                # This is a special condition where we may want to close the
                # worker if the queue stays empty
                if self.close_on_empty_queue:
                    # after we just waited, let's check the queue size again
                    if self.queue_size() == 0:
                        # if it's still empty, we should close the worker
                        logging.info("The task queue is empty. Shutting down.")
                        return

            # make this atomic so that multiple workers don't accidentally
            # grab the same job.
            with transaction.atomic():
                # If we've made it this far, we're ready to grab a new WorkItem
                # and run it!
                # Query for PENDING WorkItems, lock it for editting, and update
                # the status to RUNNING. And grab the first result
                workitem = (
                    WorkItem.objects.select_for_update(skip_locked=True)
                    .filter(status="P")
                    .filter_by_tags(self.tags)
                    .first()
                )

                # Catch race condition where no workitems are available any more.
                # If this is the case, we just restart the while loop.
                if not workitem:
                    continue

                # update the status to running before starting it so no other
                # worker tries to grab the same WorkItem
                workitem.status = "R"
                # TODO: indicate that the WorkItem is with this Worker (relationship)
                workitem.save()

            # Print out the job ID that is being ran for the user to see
            logging.info(f"Running WorkItem with id {workitem.id}")

            # now let's unpickle the WorkItem components
            fxn = cloudpickle.loads(workitem.fxn)
            args = cloudpickle.loads(workitem.args)
            kwargs = cloudpickle.loads(workitem.kwargs)

            # Try running the WorkItem
            try:
                result = fxn(*args, **kwargs)
            # if it fails, we want to "capture" the error and return it
            # rather than have the Worker fail itself.
            except Exception as exception:
                traceback.print_exc()

                logging.warning(
                    "Task failed with the error shown above. \n\n"
                    "If you are unfamilar with error tracebacks and find this error "
                    "difficult to read, you can learn more about these errors "
                    "here:\n https://realpython.com/python-traceback/\n\n"
                    "Please open a new issue on our github page if you believe "
                    "this is a bug:\n https://github.com/jacksund/simmate/issues/\n\n"
                )

                # local import to prevent circular import issues
                from simmate.engine.s3_workflow import CommandNotFoundError

                # The most common error (by far) is a command-not-found issue.
                # We want to handle this separately -- whereas other exceptions
                # we just pass on to the results.
                if isinstance(exception, CommandNotFoundError):
                    logging.warning(
                        "This WorkItem failed with a 'command not found' error. "
                        "This worker is likely improperly configured or "
                        "you have a typo in your command."
                    )

                    with transaction.atomic():
                        nfailures = workitem.command_not_found_failures + 1

                        # Check if this task is problematic. If this error happened
                        # with another worker, we likely have a problematic task
                        if nfailures == 2:
                            logging.warning(
                                "This is the 2nd occurance with this task causing "
                                "a 'command not found' problem. In case this a typo "
                                "in your command, we are marking the task as CANCELLED "
                                "to prevent it from shutting down other workers."
                            )
                            workitem.status = "C"
                            workitem.save()
                            # the result will be set below

                        # Otherwise the user likely just forgot to use module load
                        else:
                            logging.info(
                                f"Resetting WorkItem {workitem.id} to 'Pending' so "
                                "another worker can retry."
                            )

                            workitem.command_not_found_failures = nfailures
                            workitem.status = "P"  # marked as PENDING to retry
                            workitem.save()
                    logging.info("Shutting down to prevent repeated issues.")
                    return

                result = exception

            # whatever the result, we need to try to pickle it now
            try:
                result_pickled = cloudpickle.dumps(result)
            # if this fails, we even want to pickle the error and return it
            except Exception as exception:
                # otherwise package the full error
                result_pickled = cloudpickle.dumps(exception)

            # our lock exists only within this transation
            with transaction.atomic():
                # requery the WorkItem to restart our lock
                workitem = WorkItem.objects.select_for_update().get(pk=workitem.pk)

                # pickle the result and update the workitem's result and status
                # !!! should I have the pickle inside of a Try?
                workitem.result_binary = result_pickled
                # mark as finished or errored depending on result value
                workitem.status = "E" if isinstance(result, Exception) else "F"
                workitem.save()

            # mark down that we've completed one WorkItem
            ntasks_finished += 1

            # Print out the job ID that was just finished for the user to see.
            logging.info("Completed WorkItem")

    def queue_size(self) -> int:
        """
        Return the approximate size of the queue.
        """
        # Count the number of WorkItem(s) that have a status of PENDING
        # !!! Should I include RUNNING in the count? If so I do that with...
        #   from django.db.models import Q
        #   ...filter(Q(status="P") | Q(status="R"))
        queue_size = (
            WorkItem.objects.filter(status="P").filter_by_tags(self.tags).count()
        )
        return queue_size

    @classmethod
    def run_singleflow_worker(cls):
        """
        A convenience method for running a worker that...
        1. shuts down immediately if the queue is empty
        2. shuts down after a single workflow run

        Because this type of worker is frequently used for HPC clusters, we
        make a convenience method for it.
        """
        worker = cls(
            nitems_max=1,
            close_on_empty_queue=True,
            tags=["simmate"],
        )
        worker.start()



# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .base import Selector
from .truncate import TruncatedSelection
from .tournament import TournamentSelection



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.matproj import (
    Relaxation__Vasp__Matproj,
)


class Relaxation__Vasp__MatprojPbesol(Relaxation__Vasp__Matproj):
    """
    This task is a reimplementation of pymatgen's
    [MPRelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPRelaxSet).

    Runs a VASP geometry optimization using Materials Project settings with
    added settings for PBEsol.

    Materials Project settings are often considered the minimum-required
    quality for publication and is sufficient for most applications. If you are
    looking at one structure in detail (for electronic, vibrational, and other
    properties), you should still test for convergence using higher-quality
    settings.
    """

    description_doc_short = "based on pymatgen's MPMetalRelaxSet"

    # Tell VASP to use PBEsol instead of base PBE
    _incar_updates = dict(GGA="PS")



# -*- coding: utf-8 -*-

from pathlib import Path

from pymatgen.io.vasp.inputs import Kpoints

from simmate.apps.materials_project.workflows.static_energy.matproj import (
    StaticEnergy__Vasp__Matproj,
)
from simmate.apps.vasp.inputs import Incar, Poscar, Potcar
from simmate.apps.vasp.workflows.electronic_structure import get_hse_kpoints
from simmate.toolkit import Structure


class VaspDensityOfStates(StaticEnergy__Vasp__Matproj):
    """
    A base class for density of states (DOS) calculations. This is not meant
    to be used directly but instead should be inherited from.

    This is also a non self-consistent field (non-SCF) calculation and thus uses
    the a fixed charge density from a previous static energy calculation.
    """

    use_previous_directory = True
    required_files = StaticEnergy__Vasp__Matproj.required_files + ["CHGCAR"]

    @classmethod
    def setup(cls, directory: Path, structure: Structure, **kwargs):
        # run cleaning and standardizing on structure (based on class attributes)
        structure_cleaned = cls._get_clean_structure(structure, **kwargs)

        # write the poscar file
        Poscar.to_file(structure_cleaned, directory / "POSCAR")

        # Combine our base incar settings with those of our parallelization settings
        # and then write the incar file
        incar = Incar(**cls.incar)
        incar.to_file(
            filename=directory / "INCAR",
            structure=structure_cleaned,
        )

        ##############
        # we need to find the high-symmetry Kpt path. Note that all of this
        # functionality will be moved to the KptPath class and then extended to
        # vasp.inputs.kpoints class. Until those classes are ready, we just use
        # pymatgen here.
        if incar.get("LHFCALC", False):
            kpoints = get_hse_kpoints(structure_cleaned, mode="uniform")
        elif cls.kpoints and ("KSPACING" not in cls.incar):
            Kpoints.to_file(
                structure_cleaned,
                cls.kpoints,
                directory / "KPOINTS",
            )
        else:
            kpoints = None
        if kpoints:
            kpoints.write_file(directory / "KPOINTS")
        ##############

        # write the POTCAR file
        Potcar.to_file_from_type(
            structure_cleaned.composition.elements,
            cls.functional,
            directory / "POTCAR",
            cls.potcar_mappings,
        )



# -*- coding: utf-8 -*-

from simmate.apps.vasp.workflows.relaxation.quality04 import Relaxation__Vasp__Quality04


class StaticEnergy__Vasp__Quality04(Relaxation__Vasp__Quality04):
    """
    Runs a rough VASP static energy calculation.

    `Quality 04` relates to our ranking of relaxation qualities, where this
    calculation uses the same settings as the Quality04Relaxation.

    Note, even though this is currently our highest quality preset, these
    settings are still only suitable for high-throughput calculations or massive
    supercells. Settings are still below MIT and Materials Project quality.

    Most commonly, this is used in evolutionary searches (for structure
    prediction). We recommend instead using the relaxation/staged workflow,
    which uses this calculation as the sixth and final step -- after a series
    of rough relaxations are done.
    """

    _incar_updates = dict(
        ALGO="Normal",
        IBRION=-1,  # (optional) locks everything between ionic steps
        NSW=0,  # this is the main static energy setting
        ISMEAR=-5,  # was 0 for non-metals and 1 for metals
        SIGMA=0.05,  # was 0.05 for non-metals and 0.06 for metals
        multiple_keywords__smart_ismear="__remove__",
    )



# -*- coding: utf-8 -*-

from simmate.apps.bader.configuration import test_config as test_bader_config
from simmate.apps.vasp.configuration import test_config as test_vasp_config
from simmate.configuration.utilities import check_app_reg, show_test_results


def test_config(run_calcs: bool = False):
    """
    Checks to see if Material Project App is configured correctly
    """

    # 1 - check that Simmate app is registered
    is_registered = check_app_reg("simmate.apps.configs.BadelfConfig")

    # 2 - check app dependancies
    has_vasp = test_vasp_config()
    has_bader = test_bader_config()

    # 3 - run some sample workflows
    if run_calcs:
        raise NotImplementedError("This test has not been added yet.")

    # 4 - read out result of all tests
    passed_all = bool(is_registered and has_bader and has_vasp)
    show_test_results("BadELF", passed_all)



# -*- coding: utf-8 -*-

from simmate.toolkit.transformations.base import Transformation


class CoordinateOrderedPerturbation(Transformation):
    """
    (NOT YET IMPLEMENTED: THIS IS A PLACEHOLDER)

    Known as "coordinate mutation" in USPEX. Here, site locations are mutated
    where sites with lower order have higher preference for mutation
        https://uspex-team.org/static/file/USPEX-LargeComplexSystems-2010.pdf

    From one of their papers:
    > Coordinate mutation was found [2] to be ineffective, because “blind”
      displacement of the atoms is much more likely to decrease the quality
      of a structure than to increase it.
    """

    io_scale = "one_to_one"
    ninput = 1
    allow_parallel = False
